"""AccuSleePy manual scoring GUI.

Icon sources:
    Arkinasi, https://www.flaticon.com/authors/arkinasi
    kendis lasman, https://www.flaticon.com/packs/ui-79
"""

import copy
import os
import time
from dataclasses import dataclass
from functools import partial
from types import SimpleNamespace

import matplotlib.pyplot as plt
import numpy as np
from PySide6.QtCore import (
    QKeyCombination,
    QRect,
    Qt,
    QUrl,
)
from PySide6.QtGui import (
    QCloseEvent,
    QKeySequence,
    QShortcut,
)
from PySide6.QtWidgets import (
    QDialog,
    QMessageBox,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

from accusleepy.constants import UNDEFINED_LABEL
from accusleepy.fileio import EMGFilter, load_config, save_labels
from accusleepy.gui.mplwidget import resample_x_ticks
from accusleepy.gui.viewer_window import Ui_ViewerWindow
from accusleepy.signal_processing import create_spectrogram, get_emg_power

# colormap for displaying brain state labels
# the first entry represents the "undefined" state
# the other entries are the digits in "keyboard" order (1234567890)
LABEL_CMAP = np.concatenate(
    [np.array([[0, 0, 0, 0]]), plt.colormaps["tab10"](range(10))], axis=0
)
# relative path to user manual text file
USER_MANUAL_FILE = os.path.normpath(r"text/manual_scoring_guide.md")

# constants used by callback functions
# label formats
DISPLAY_FORMAT = "display"
DIGIT_FORMAT = "digit"
# offset changes
OFFSET_UP = "up"
OFFSET_DOWN = "down"
OFFSET_INCREMENTS = {OFFSET_UP: 0.02, OFFSET_DOWN: -0.02}
# changes to number of epochs
DIRECTION_PLUS = "plus"
DIRECTION_MINUS = "minus"
# changes to selected epoch
DIRECTION_LEFT = "left"
DIRECTION_RIGHT = "right"
# zoom directions
ZOOM_IN = "in"
ZOOM_OUT = "out"
ZOOM_RESET = "reset"
SIGNAL_ZOOM_FACTORS = {ZOOM_IN: 1.08, ZOOM_OUT: 0.95}
# signal names
EEG_SIGNAL = "eeg"
EMG_SIGNAL = "emg"
# spectrogram color changes
BRIGHTER = "brighter"
DIMMER = "dimmer"
# next epoch target
DIFFERENT_STATE = "different"
UNDEFINED_STATE = "undefined"
# how far from the edge of the upper plot the marker should be
# before starting to scroll again - must be in (0, 0.5)
SCROLL_BOUNDARY = 0.35
# max number of sequential undo actions allowed
UNDO_LIMIT = 1000
# brightness scaling factors for the spectrogram
BRIGHTER_SCALE_FACTOR = 0.96
DIMMER_SCALE_FACTOR = 1.07
# zoom factor for upper plots - larger values = bigger changes
ZOOM_FACTOR = 0.1
# rate limit for zoom events triggered by scrolling
MAX_SCROLL_EVENTS_PER_SEC = 24
# tip to show when EEG/EMG plot moves a lot
RESET_TIP = "click plot 2x to reset"


@dataclass
class StateChange:
    """Information about an event when brain state labels were changed"""

    previous_labels: np.array  # old brain state labels
    new_labels: np.array  # new brain state labels
    epoch: int  # first epoch affected


class ManualScoringWindow(QDialog):
    """AccuSleePy manual scoring GUI"""

    def __init__(
        self,
        eeg: np.array,
        emg: np.array,
        label_file: str,
        labels: np.array,
        confidence_scores: np.array,
        sampling_rate: int | float,
        epoch_length: int | float,
        emg_filter: EMGFilter,
    ):
        """Initialize the manual scoring window

        :param eeg: EEG signal
        :param emg: EMG signal
        :param label_file: filename for labels
        :param labels: brain state labels
        :param confidence_scores: confidence scores
        :param sampling_rate: sampling rate, in Hz
        :param epoch_length: epoch length, in seconds
        :param emg_filter: EMG filter parameters
        """
        super().__init__()

        self.label_file = label_file
        self.eeg = eeg
        self.emg = emg
        self.labels = labels
        self.confidence_scores = confidence_scores
        self.sampling_rate = sampling_rate
        self.epoch_length = epoch_length
        self.emg_filter = emg_filter

        self.n_epochs = len(self.labels)

        # initialize the UI
        self.ui = Ui_ViewerWindow()
        self.ui.setupUi(self)
        self.setWindowTitle("AccuSleePy manual scoring window")

        config = load_config()
        self.brain_state_set = config.brain_state_set
        self.epochs_to_show = config.epochs_to_show
        self.autoscroll_state = config.autoscroll_state

        # find the set of y-axis locations of valid brain state labels
        self.label_display_options = convert_labels(
            np.array([b.digit for b in self.brain_state_set.brain_states]),
            style=DISPLAY_FORMAT,
        )
        self.smallest_display_label = np.min(self.label_display_options)

        self.ui.upperfigure.epoch_length = self.epoch_length
        self.ui.lowerfigure.epoch_length = self.epoch_length

        # get EEG spectrogram and its frequency axis
        spectrogram, spectrogram_frequencies = create_spectrogram(
            self.eeg, self.sampling_rate, self.epoch_length
        )

        # calculate RMS of EMG for each epoch and apply a ceiling
        self.upper_emg = create_upper_emg_signal(
            self.emg, self.sampling_rate, self.epoch_length, self.emg_filter
        )

        # center and scale the EEG and EMG signals to fit the display
        self.eeg, self.emg = transform_eeg_emg(self.eeg, self.emg)
        # only a portion will be displayed in the lower plots
        # we'll store this so that it's easy to modify
        self.eeg_shown = None
        self.emg_shown = None

        # convert labels to "display" format and make an image to display them
        self.display_labels = convert_labels(self.labels, DISPLAY_FORMAT)
        self.label_img = create_label_img(
            self.display_labels, self.label_display_options
        )
        # same sort of thing for confidence scores
        self.confidence_img = create_confidence_img(self.confidence_scores)

        # history of changes to the brain state labels
        self.history = list()
        # index of the change "ahead" of the current state
        # i.e., which change will be applied by a "redo" action
        self.history_index = 0

        # set up both figures
        self.ui.upperfigure.setup_upper_figure(
            self.n_epochs,
            self.label_img,
            self.confidence_scores,
            self.confidence_img,
            spectrogram,
            spectrogram_frequencies,
            self.upper_emg,
            self.epochs_to_show,
            self.label_display_options,
            self.brain_state_set,
            self.roi_callback,
        )
        self.ui.lowerfigure.setup_lower_figure(
            self.label_img,
            self.sampling_rate,
            self.epochs_to_show,
            self.brain_state_set,
            self.label_display_options,
        )

        # initialize values that can be changed by user input
        self.epoch = 0
        self.upper_left_epoch = 0
        self.upper_right_epoch = self.n_epochs - 1
        self.lower_left_epoch = 0
        self.lower_right_epoch = self.epochs_to_show - 1
        self.eeg_signal_scale_factor = 1
        self.emg_signal_scale_factor = 1
        self.eeg_signal_offset = 0
        self.emg_signal_offset = 0
        self.roi_brain_state = 0
        self.label_roi_mode = False
        self.epoch_markers_shown = False

        # set autoscroll state and epochs to show, based on defaults
        self.ui.autoscroll.setChecked(self.autoscroll_state)
        self.ui.shownepochslabel.setText(str(self.epochs_to_show))

        # keep track of save state to warn user when they quit
        self.last_saved_labels = copy.deepcopy(self.labels)

        # no need to show the tip about the lower axes yet
        self.ui.reset_tip_label.setText("")

        # populate the lower figure
        self.update_lower_figure()

        # user input: keyboard shortcuts
        keypress_right = QShortcut(QKeySequence(Qt.Key.Key_Right), self)
        keypress_right.activated.connect(partial(self.shift_epoch, DIRECTION_RIGHT))

        keypress_left = QShortcut(QKeySequence(Qt.Key.Key_Left), self)
        keypress_left.activated.connect(partial(self.shift_epoch, DIRECTION_LEFT))

        keypress_zoom_in_x = list()
        for zoom_key in [Qt.Key.Key_Plus, Qt.Key.Key_Equal]:
            keypress_zoom_in_x.append(QShortcut(QKeySequence(zoom_key), self))
            keypress_zoom_in_x[-1].activated.connect(partial(self.zoom_x, ZOOM_IN))

        keypress_zoom_out_x = QShortcut(QKeySequence(Qt.Key.Key_Minus), self)
        keypress_zoom_out_x.activated.connect(partial(self.zoom_x, ZOOM_OUT))

        keypress_quit = QShortcut(
            QKeySequence(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_W)),
            self,
        )
        keypress_quit.activated.connect(self.close)

        keypress_save = QShortcut(
            QKeySequence(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_S)),
            self,
        )
        keypress_save.activated.connect(self.save)

        keypress_modify_label = list()
        keypress_roi = list()
        digit_key_label_pairs = [
            (Qt.Key[f"Key_{brain_state.digit}"], brain_state.digit)
            for brain_state in self.brain_state_set.brain_states
        ] + [(Qt.Key.Key_Backspace, UNDEFINED_LABEL)]

        for digit_key, digit_label in digit_key_label_pairs:
            keypress_modify_label.append(
                QShortcut(
                    QKeySequence(digit_key),
                    self,
                )
            )
            keypress_modify_label[-1].activated.connect(
                partial(self.modify_current_epoch_label, digit_label)
            )
            keypress_roi.append(
                QShortcut(
                    QKeySequence(
                        QKeyCombination(
                            Qt.Modifier.SHIFT,
                            digit_key,
                        )
                    ),
                    self,
                )
            )
            keypress_roi[-1].activated.connect(
                partial(self.enter_label_roi_mode, digit_label)
            )

        keypress_esc = QShortcut(QKeySequence(Qt.Key.Key_Escape), self)
        keypress_esc.activated.connect(self.exit_label_roi_mode)

        keypress_space = QShortcut(QKeySequence(Qt.Key.Key_Space), self)
        keypress_space.activated.connect(
            partial(self.jump_to_next_state, DIRECTION_RIGHT, DIFFERENT_STATE)
        )
        keypress_shift_right = QShortcut(
            QKeySequence(
                QKeyCombination(
                    Qt.Modifier.SHIFT,
                    Qt.Key.Key_Right,
                )
            ),
            self,
        )
        keypress_shift_right.activated.connect(
            partial(self.jump_to_next_state, DIRECTION_RIGHT, DIFFERENT_STATE)
        )
        keypress_shift_left = QShortcut(
            QKeySequence(
                QKeyCombination(
                    Qt.Modifier.SHIFT,
                    Qt.Key.Key_Left,
                )
            ),
            self,
        )
        keypress_shift_left.activated.connect(
            partial(self.jump_to_next_state, DIRECTION_LEFT, DIFFERENT_STATE)
        )
        keypress_ctrl_right = QShortcut(
            QKeySequence(
                QKeyCombination(
                    Qt.Modifier.CTRL,
                    Qt.Key.Key_Right,
                )
            ),
            self,
        )
        keypress_ctrl_right.activated.connect(
            partial(self.jump_to_next_state, DIRECTION_RIGHT, UNDEFINED_STATE)
        )
        keypress_ctrl_left = QShortcut(
            QKeySequence(
                QKeyCombination(
                    Qt.Modifier.CTRL,
                    Qt.Key.Key_Left,
                )
            ),
            self,
        )
        keypress_ctrl_left.activated.connect(
            partial(self.jump_to_next_state, DIRECTION_LEFT, UNDEFINED_STATE)
        )

        keypress_undo = QShortcut(
            QKeySequence(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Z)),
            self,
        )
        keypress_undo.activated.connect(self.undo)
        keypress_redo = QShortcut(
            QKeySequence(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Y)),
            self,
        )
        keypress_redo.activated.connect(self.redo)

        # user input: mouse events
        self.ui.upperfigure.canvas.mpl_connect("button_press_event", self.click_to_jump)
        self.ui.upperfigure.canvas.mpl_connect(
            "scroll_event", partial(self.scroll_zoom, "upper")
        )
        self.ui.lowerfigure.canvas.mpl_connect(
            "scroll_event", partial(self.scroll_zoom, "lower")
        )
        self.ui.lowerfigure.canvas.mpl_connect(
            "button_press_event", self.lower_plot_click
        )
        self.ui.lowerfigure.canvas.mpl_connect(
            "button_release_event", self.lower_plot_click
        )
        self.ui.lowerfigure.canvas.mpl_connect(
            "motion_notify_event", self.lower_plot_move
        )
        self.now_zooming = False  # impose timeout on zoom events
        self.drag_axes = None  # axes experiencing a drag event
        self.last_y = None  # most recent mouse y position

        # user input: buttons
        self.ui.savebutton.clicked.connect(self.save)
        self.ui.xzoomin.clicked.connect(partial(self.zoom_x, ZOOM_IN))
        self.ui.xzoomout.clicked.connect(partial(self.zoom_x, ZOOM_OUT))
        self.ui.xzoomreset.clicked.connect(partial(self.zoom_x, ZOOM_RESET))
        self.ui.autoscroll.stateChanged.connect(self.update_autoscroll_state)
        self.ui.eegzoomin.clicked.connect(
            partial(self.update_signal_zoom, EEG_SIGNAL, ZOOM_IN)
        )
        self.ui.eegzoomout.clicked.connect(
            partial(self.update_signal_zoom, EEG_SIGNAL, ZOOM_OUT)
        )
        self.ui.emgzoomin.clicked.connect(
            partial(self.update_signal_zoom, EMG_SIGNAL, ZOOM_IN)
        )
        self.ui.emgzoomout.clicked.connect(
            partial(self.update_signal_zoom, EMG_SIGNAL, ZOOM_OUT)
        )
        self.ui.eegshiftup.clicked.connect(
            partial(self.update_signal_offset, EEG_SIGNAL, OFFSET_UP)
        )
        self.ui.eegshiftdown.clicked.connect(
            partial(self.update_signal_offset, EEG_SIGNAL, OFFSET_DOWN)
        )
        self.ui.emgshiftup.clicked.connect(
            partial(self.update_signal_offset, EMG_SIGNAL, OFFSET_UP)
        )
        self.ui.emgshiftdown.clicked.connect(
            partial(self.update_signal_offset, EMG_SIGNAL, OFFSET_DOWN)
        )
        self.ui.shownepochsplus.clicked.connect(
            partial(self.update_epochs_shown, DIRECTION_PLUS)
        )
        self.ui.shownepochsminus.clicked.connect(
            partial(self.update_epochs_shown, DIRECTION_MINUS)
        )
        self.ui.specbrighter.clicked.connect(
            partial(self.update_spectrogram_brightness, BRIGHTER)
        )
        self.ui.specdimmer.clicked.connect(
            partial(self.update_spectrogram_brightness, DIMMER)
        )
        self.ui.helpbutton.clicked.connect(self.show_user_manual)

        self.show()

    def add_to_history(self, state_change: StateChange) -> None:
        """Add an event to the history of changes to brain state labels

        This allows the user to undo / redo changes to the brain state
        labels by navigating backwards or forwards through a list of
        changes that have been made. If one or more changes are undone,
        and then a new change is performed, it will not be possible to
        redo the changes that were undone. At most UNDO_LIMIT changes
        are stored.

        :param state_change: description of the change to the labels
        """
        # if history is empty
        if len(self.history) == 0:
            self.history.append(state_change)
            self.history_index = 1
            return
        # if we are not at the end of the history
        if self.history_index < len(self.history):
            # remove events after the most recent one
            self.history = self.history[: self.history_index]
            self.history.append(state_change)
            self.history_index += 1
        else:
            self.history.append(state_change)
            # if this would make the history list too long
            if self.history_index == UNDO_LIMIT:
                # remove the oldest entry
                self.history = self.history[1:]
            else:
                self.history_index += 1

    def force_modify_labels(self, epoch: int, new_labels: np.array) -> None:
        """Change brain state labels for an undo/redo action"""
        # make the change
        self.labels[epoch : epoch + len(new_labels)] = new_labels
        self.display_labels = convert_labels(
            self.labels,
            style=DISPLAY_FORMAT,
        )
        self.label_img = create_label_img(
            self.display_labels, self.label_display_options
        )

    def redo(self) -> None:
        """Redo the last change to brain state labels that was undone"""
        # if there are no events to redo
        if self.history_index == len(self.history):
            return
        # make the change
        state_change = self.history[self.history_index]
        self.force_modify_labels(
            epoch=state_change.epoch, new_labels=state_change.new_labels
        )
        # update history index
        self.history_index += 1
        # shift the cursor
        simulated_click = SimpleNamespace(
            **{"xdata": state_change.epoch, "inaxes": None}
        )
        self.click_to_jump(simulated_click)

    def undo(self) -> None:
        """Undo the last change to the brain state labels"""
        # if there are no events to undo
        if self.history_index == 0:
            return
        # make the change
        state_change = self.history[self.history_index - 1]
        self.force_modify_labels(
            epoch=state_change.epoch, new_labels=state_change.previous_labels
        )
        # update history index
        self.history_index -= 1
        # shift the cursor
        simulated_click = SimpleNamespace(
            **{"xdata": state_change.epoch, "inaxes": None}
        )
        self.click_to_jump(simulated_click)

    def closeEvent(self, event: QCloseEvent) -> None:
        """Check if there are unsaved changes before closing"""
        if not np.array_equal(self.labels, self.last_saved_labels):
            result = QMessageBox.question(
                self,
                "Unsaved changes",
                "You have unsaved changes. Really quit?",
                QMessageBox.Yes | QMessageBox.No,
            )
            if result == QMessageBox.Yes:
                event.accept()
            else:
                event.ignore()

    def show_user_manual(self) -> None:
        """Show a popup window with the user manual"""
        self.popup = QWidget()
        self.popup_vlayout = QVBoxLayout(self.popup)
        self.guide_textbox = QTextBrowser(self.popup)
        self.popup_vlayout.addWidget(self.guide_textbox)

        url = QUrl.fromLocalFile(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), USER_MANUAL_FILE)
        )
        self.guide_textbox.setSource(url)
        self.guide_textbox.setOpenLinks(False)

        self.popup.setGeometry(QRect(100, 100, 830, 600))
        self.popup.show()

    def jump_to_next_state(self, direction: str, target: str) -> None:
        """Jump to epoch based on a target brain state

        This allows the user to jump to the next epoch in a given direction
        (left or right) that has a given state (undefined, or different from
        the current epoch). It's useful for reviewing state transitions or
        locating unlabeled epochs.

        :param direction: left or right
        :param target: different or undefined
        """
        # create a simulated click so we can reuse click_to_jump
        simulated_click = SimpleNamespace(**{"xdata": self.epoch, "inaxes": None})
        if direction == DIRECTION_RIGHT:
            if target == DIFFERENT_STATE:
                matches = np.where(
                    self.labels[self.epoch + 1 :] != self.labels[self.epoch]
                )[0]
            else:
                matches = np.where(self.labels[self.epoch + 1 :] == UNDEFINED_LABEL)[0]
            if matches.size > 0:
                simulated_click.xdata = int(matches[0]) + 1 + self.epoch
        else:
            if target == DIFFERENT_STATE:
                matches = np.where(
                    self.labels[: self.epoch] != self.labels[self.epoch]
                )[0]
            else:
                matches = np.where(self.labels[: self.epoch] == UNDEFINED_LABEL)[0]
            if matches.size > 0:
                simulated_click.xdata = int(matches[-1])
        self.click_to_jump(simulated_click)

    def roi_callback(self, eclick, erelease) -> None:
        """Callback for ROI labeling widget

        This is called by the RectangleSelector widget when the user finishes
        drawing an ROI. It sets a range of epochs to the desired brain state.
        The function signature is required to have this format.
        """
        # get range of epochs affected
        epoch_start = int(np.ceil(eclick.xdata))
        epoch_end = int(np.floor(erelease.xdata)) + 1

        previous_labels = copy.deepcopy(self.labels[epoch_start:epoch_end])
        new_labels = np.ones(epoch_end - epoch_start).astype(int) * self.roi_brain_state

        # if something changed, track the change
        if not np.array_equal(previous_labels, new_labels):
            self.add_to_history(
                StateChange(
                    previous_labels=previous_labels,
                    new_labels=new_labels,
                    epoch=epoch_start,
                )
            )
        # make the change
        self.labels[epoch_start:epoch_end] = self.roi_brain_state
        self.display_labels = convert_labels(
            self.labels,
            style=DISPLAY_FORMAT,
        )
        self.label_img = create_label_img(
            self.display_labels, self.label_display_options
        )
        # update the plots
        self.update_figures()
        self.exit_label_roi_mode()

    def exit_label_roi_mode(self) -> None:
        """Restore the normal GUI state after an ROI is drawn"""
        self.ui.upperfigure.roi.set_active(False)
        self.ui.upperfigure.roi.set_visible(False)
        self.ui.upperfigure.roi.update()
        self.label_roi_mode = False
        self.ui.upperfigure.editing_patch.set_visible(False)

    def enter_label_roi_mode(self, brain_state: int) -> None:
        """Enter ROI drawing mode

        In this mode, a user can draw an ROI on the upper brain state label
        image to set a range of epochs to a new brain state.

        :param brain_state: new brain state to set
        """
        self.label_roi_mode = True
        self.roi_brain_state = brain_state
        self.ui.upperfigure.roi_patch.set(
            facecolor=LABEL_CMAP[
                convert_labels(np.array([brain_state]), DISPLAY_FORMAT)
            ]
        )
        self.ui.upperfigure.editing_patch.set_visible(True)
        self.ui.upperfigure.canvas.draw()
        self.ui.upperfigure.roi.set_active(True)

    def save(self) -> None:
        """Save brain state labels to file"""
        save_labels(self.labels, self.label_file)
        self.last_saved_labels = copy.deepcopy(self.labels)

    def update_spectrogram_brightness(self, direction: str) -> None:
        """Modify spectrogram color range based on button press

        :param direction: brighter or dimmer
        """
        vmin, vmax = self.ui.upperfigure.spec_ref.get_clim()
        if direction == BRIGHTER:
            self.ui.upperfigure.spec_ref.set(clim=(vmin, vmax * BRIGHTER_SCALE_FACTOR))
        else:
            self.ui.upperfigure.spec_ref.set(clim=(vmin, vmax * DIMMER_SCALE_FACTOR))
        self.ui.upperfigure.canvas.draw()

    def update_epochs_shown(self, direction: str) -> None:
        """Change the number of epochs shown based on button press

        The user can change the number of epochs shown in the lower figure
        via button presses. This requires extensive changes to both figures.
        The number of epochs can only change in increments of 2 and should
        always be an odd number >= 3.

        :param direction: plus or minus
        """
        # if we are near the beginning or end of the recording, we need
        # to change the epoch range differently.
        if direction == DIRECTION_PLUS:
            self.epochs_to_show += 2
            if self.lower_left_epoch == 0:
                self.lower_right_epoch += 2
            elif self.lower_right_epoch == self.n_epochs - 1:
                self.lower_left_epoch -= 2
            else:
                self.lower_left_epoch -= 1
                self.lower_right_epoch += 1
        else:
            if self.epochs_to_show > 3:
                self.epochs_to_show -= 2
                if self.lower_left_epoch == 0:
                    self.lower_right_epoch -= 2
                elif self.lower_right_epoch == self.n_epochs - 1:
                    self.lower_left_epoch += 2
                else:
                    self.lower_left_epoch += 1
                    self.lower_right_epoch -= 1

        self.ui.shownepochslabel.setText(str(self.epochs_to_show))

        # rebuild lower figure from scratch
        self.ui.lowerfigure.canvas.figure.clf()
        self.ui.lowerfigure.setup_lower_figure(
            self.label_img,
            self.sampling_rate,
            self.epochs_to_show,
            self.brain_state_set,
            self.label_display_options,
        )
        self.update_figures()
        # upper plot x limits might need to change
        self.zoom_x(direction=None)

    def _refresh_eeg_line(self) -> None:
        """Recompute and update the displayed EEG line."""
        adjusted = (
            self.eeg_shown * self.eeg_signal_scale_factor + self.eeg_signal_offset
        )
        self.ui.lowerfigure.eeg_line.set_ydata(adjusted)

    def _refresh_emg_line(self) -> None:
        """Recompute and update the displayed EMG line."""
        adjusted = (
            self.emg_shown * self.emg_signal_scale_factor + self.emg_signal_offset
        )
        self.ui.lowerfigure.emg_line.set_ydata(adjusted)

    def update_signal_offset(self, signal: str, direction: str) -> None:
        """Shift EEG or EMG up or down

        :param signal: eeg or emg
        :param direction: up or down
        """
        if signal == EEG_SIGNAL:
            self.eeg_signal_offset += OFFSET_INCREMENTS[direction]
            self._refresh_eeg_line()
        else:
            self.emg_signal_offset += OFFSET_INCREMENTS[direction]
            self._refresh_emg_line()
        self.ui.lowerfigure.canvas.draw()

    def update_signal_zoom(self, signal: str, direction: str) -> None:
        """Zoom EEG or EMG y-axis

        :param signal: eeg or emg
        :param direction: in or out
        """
        if signal == EEG_SIGNAL:
            self.eeg_signal_scale_factor *= SIGNAL_ZOOM_FACTORS[direction]
        else:
            self.emg_signal_scale_factor *= SIGNAL_ZOOM_FACTORS[direction]
        self.update_lower_figure()

    def update_autoscroll_state(self, checked) -> None:
        """Toggle autoscroll behavior

        If autoscroll is enabled, setting the brain state of the current epoch
        via a keypress will advance to the next epoch.

        :param checked: state of the checkbox
        """
        self.autoscroll_state = checked

    def adjust_upper_figure_x_limits(self) -> None:
        """Update the x-axis limits of the upper figure subplots"""
        for i in [0, 1, 2, 4]:
            self.ui.upperfigure.canvas.axes[i].set_xlim(
                (self.upper_left_epoch - 0.5, self.upper_right_epoch + 0.5)
            )
        self.ui.upperfigure.canvas.axes[3].set_xlim(
            (self.upper_left_epoch, self.upper_right_epoch + 1)
        )

    def zoom_x(self, direction: str | None) -> None:
        """Change upper figure x-axis zoom level

        :param direction: in, out, reset, or None
        """
        self.upper_left_epoch, self.upper_right_epoch = find_new_x_limits(
            direction=direction,
            left_epoch=self.upper_left_epoch,
            right_epoch=self.upper_right_epoch,
            min_n_shown=self.epochs_to_show,
            total_epochs=self.n_epochs,
            selected_epoch=self.epoch,
        )
        self.adjust_upper_figure_x_limits()
        self.ui.upperfigure.canvas.draw()

    def modify_current_epoch_label(self, digit: int) -> None:
        """Change the current epoch's brain state label

        :param digit: new brain state label in "digit" format
        """
        previous_label = self.labels[self.epoch]
        self.labels[self.epoch] = digit
        # if something changed, track the change
        if previous_label != digit:
            self.add_to_history(
                StateChange(
                    previous_labels=np.array([previous_label]),
                    new_labels=np.array([digit]),
                    epoch=self.epoch,
                )
            )

        # make the change
        display_label = convert_labels(
            np.array([digit]),
            style=DISPLAY_FORMAT,
        )[0]
        self.display_labels[self.epoch] = display_label
        # update the label image
        if display_label == 0:
            self.label_img[:, self.epoch] = np.array([0, 0, 0, 1])
        else:
            self.label_img[:, self.epoch, :] = 1
            self.label_img[
                display_label - self.smallest_display_label, self.epoch, :
            ] = LABEL_CMAP[display_label]
        # autoscroll, if that is enabled
        if self.autoscroll_state and self.epoch < self.n_epochs - 1:
            self.shift_epoch(DIRECTION_RIGHT)  # this calls update_figures()
        else:
            self.update_figures()

    def shift_epoch(self, direction: str) -> None:
        """Set the current epoch one step forward or backward

        When the user presses the left or right arrow key, the previous
        or next epoch will be selected. There are a variety of edge cases
        that need to be handled separately for the upper and lower figures.

        :param direction: left or right
        """
        shift_amount = {DIRECTION_LEFT: -1, DIRECTION_RIGHT: 1}[direction]
        # prevent movement outside the data range
        if not (0 <= (self.epoch + shift_amount) < self.n_epochs):
            return

        # shift to new epoch
        self.epoch = self.epoch + shift_amount

        # update upper plot if needed
        upper_epochs_shown = self.upper_right_epoch - self.upper_left_epoch + 1
        if (
            (
                self.epoch
                > self.upper_left_epoch + (1 - SCROLL_BOUNDARY) * upper_epochs_shown
                or self.epoch + (self.epochs_to_show - 1) / 2 > self.upper_right_epoch
            )
            and self.upper_right_epoch < (self.n_epochs - 1)
            and direction == DIRECTION_RIGHT
        ):
            self.upper_left_epoch += 1
            self.upper_right_epoch += 1
            self.adjust_upper_figure_x_limits()
        elif (
            (
                self.epoch
                < self.upper_left_epoch + SCROLL_BOUNDARY * upper_epochs_shown
                or self.epoch - (self.epochs_to_show - 1) / 2 < self.upper_left_epoch
            )
            and self.upper_left_epoch > 0
            and direction == DIRECTION_LEFT
        ):
            self.upper_left_epoch -= 1
            self.upper_right_epoch -= 1
            self.adjust_upper_figure_x_limits()

        # update parts of lower plot
        old_window_center = round((self.epochs_to_show - 1) / 2) + self.lower_left_epoch
        # change the window bounds if needed
        if self.epoch < old_window_center and self.lower_left_epoch > 0:
            self.lower_left_epoch -= 1
            self.lower_right_epoch -= 1
        elif (
            self.epoch > old_window_center
            and self.lower_right_epoch < self.n_epochs - 1
        ):
            self.lower_left_epoch += 1
            self.lower_right_epoch += 1

        self.update_figures()

    def update_upper_marker(self) -> None:
        """Update location of the upper figure's epoch marker"""
        epoch_padding = round((self.epochs_to_show - 1) / 2)
        if self.epoch - epoch_padding < 0:
            left_edge = 0
            right_edge = self.epochs_to_show - 1
        elif self.epoch + epoch_padding > self.n_epochs - 1:
            right_edge = self.n_epochs - 1
            left_edge = self.n_epochs - self.epochs_to_show
        else:
            left_edge = self.epoch - epoch_padding
            right_edge = self.epoch + epoch_padding
        self.ui.upperfigure.upper_marker[0].set_xdata(
            [
                left_edge - 0.5,
                right_edge + 0.5,
            ]
        )
        self.ui.upperfigure.upper_marker[1].set_xdata([self.epoch])

    def update_lower_epoch_marker(self) -> None:
        """Update location of the lower figure's epoch marker"""
        marker_left = (
            (self.epoch - self.lower_left_epoch)
            * self.epoch_length
            * self.sampling_rate
        )
        marker_right = (
            (1 + self.epoch - self.lower_left_epoch)
            * self.epoch_length
            * self.sampling_rate
        )
        self.ui.lowerfigure.top_marker[0].set_xdata([marker_left, marker_left])
        self.ui.lowerfigure.top_marker[1].set_xdata([marker_left, marker_right])
        self.ui.lowerfigure.top_marker[2].set_xdata([marker_right, marker_right])
        self.ui.lowerfigure.bottom_marker[0].set_xdata([marker_left, marker_left])
        self.ui.lowerfigure.bottom_marker[1].set_xdata([marker_left, marker_right])
        self.ui.lowerfigure.bottom_marker[2].set_xdata([marker_right, marker_right])

    def update_figures(self) -> None:
        """Update and redraw both figures"""
        # upper figure
        self.update_upper_marker()
        # this step isn't always needed, but it's not too expensive
        self.ui.upperfigure.label_img_ref.set(data=self.label_img)
        self.ui.upperfigure.canvas.draw()
        # lower figure
        self.update_lower_figure()

    def update_lower_figure(self) -> None:
        """Update and redraw the lower figure"""
        # get subset of signals to plot
        first_sample = round(
            self.lower_left_epoch * self.sampling_rate * self.epoch_length
        )
        last_sample = round(
            (self.lower_right_epoch + 1) * self.sampling_rate * self.epoch_length
        )
        # store the (unmodified) eeg/emg segments to display
        self.eeg_shown = self.eeg[first_sample:last_sample]
        self.emg_shown = self.emg[first_sample:last_sample]

        self.update_lower_epoch_marker()

        # replot eeg and emg
        self._refresh_eeg_line()
        self._refresh_emg_line()

        # replot brain state
        self.ui.lowerfigure.label_img_ref.set(
            data=self.label_img[
                :, self.lower_left_epoch : (self.lower_right_epoch + 1), :
            ]
        )
        # update timestamps
        x_ticks = resample_x_ticks(
            np.arange(self.lower_left_epoch, self.lower_right_epoch + 1)
        )
        self.ui.lowerfigure.canvas.axes[1].set_xticklabels(
            [
                f"{int(x // 3600):02d}:{int(x // 60) % 60:02d}:{x % 60:05.2f}"
                for x in x_ticks * self.epoch_length
            ]
        )

        self.ui.lowerfigure.eeg_epoch_plot.set(visible=self.epoch_markers_shown)
        self.ui.lowerfigure.emg_epoch_plot.set(visible=self.epoch_markers_shown)

        self.ui.lowerfigure.canvas.draw()

    def click_to_jump(self, event) -> None:
        """Jump to a new epoch when the user clicks on the upper figure

        This is the callback for mouse clicks on the upper figure. Clicking on
        any of the subplots will jump to the nearest epoch.

        :param event: a MouseEvent containing the click data
        """
        # make sure click location is valid
        # and we are not in label ROI mode
        if event.xdata is None or self.label_roi_mode:
            return

        # get click location
        x = event.xdata
        # if it's a real click, and not one we simulated
        if event.inaxes is not None:
            # if it's on the spectrogram, we have to adjust it slightly
            # since that uses a different x-axis range
            ax_index = self.ui.upperfigure.canvas.axes.index(event.inaxes)
            if ax_index == 3:
                x -= 0.5

        # get the "zoom level" so we can preserve that
        upper_epochs_shown = self.upper_right_epoch - self.upper_left_epoch + 1
        upper_epoch_padding = round((upper_epochs_shown - 1) / 2)
        # update epoch
        self.epoch = round(np.clip(x, 0, self.n_epochs - 1))
        # update upper figure x-axis limits
        if self.epoch - upper_epoch_padding < 0:
            self.upper_left_epoch = 0
            self.upper_right_epoch = upper_epochs_shown - 1
        elif self.epoch + upper_epoch_padding > self.n_epochs - 1:
            self.upper_right_epoch = self.n_epochs - 1
            self.upper_left_epoch = self.n_epochs - upper_epochs_shown
        else:
            self.upper_left_epoch = self.epoch - upper_epoch_padding
            self.upper_right_epoch = self.epoch + upper_epoch_padding
        self.adjust_upper_figure_x_limits()

        # update lower figure x-axis range
        lower_epoch_padding = round((self.epochs_to_show - 1) / 2)
        if self.epoch - lower_epoch_padding < 0:
            self.lower_left_epoch = 0
            self.lower_right_epoch = self.epochs_to_show - 1
        elif self.epoch + lower_epoch_padding > self.n_epochs - 1:
            self.lower_right_epoch = self.n_epochs - 1
            self.lower_left_epoch = self.n_epochs - self.epochs_to_show
        else:
            self.lower_left_epoch = self.epoch - lower_epoch_padding
            self.lower_right_epoch = self.epoch + lower_epoch_padding

        self.update_figures()

    def scroll_zoom(self, canvas, event) -> None:
        """Zoom on mouse scroll events"""
        if self.now_zooming:
            return

        self.now_zooming = True
        start_time = time.time()

        direction = ZOOM_IN if event.button == "up" else ZOOM_OUT

        if canvas == "upper":
            self.zoom_x(direction=direction)
        elif event.inaxes == self.ui.lowerfigure.canvas.axes[0]:
            self.update_signal_zoom(EEG_SIGNAL, direction)
        elif event.inaxes == self.ui.lowerfigure.canvas.axes[1]:
            self.update_signal_zoom(EMG_SIGNAL, direction)

        end_time = time.time()
        time_elapsed = end_time - start_time
        if time_elapsed < 1 / MAX_SCROLL_EVENTS_PER_SEC:
            time.sleep(1 / MAX_SCROLL_EVENTS_PER_SEC - time_elapsed)
        self.now_zooming = False

    def lower_plot_click(self, event):
        """Triggered when the lower plots are clicked"""
        # right click swaps epoch marker state
        if event.button == 3 and event.name == "button_press_event":
            self.epoch_markers_shown = not self.epoch_markers_shown
            self.ui.lowerfigure.eeg_epoch_plot.set(visible=self.epoch_markers_shown)
            self.ui.lowerfigure.emg_epoch_plot.set(visible=self.epoch_markers_shown)
            self.ui.lowerfigure.canvas.draw()
            return

        # only process LMB clicks on the EEG/EMG axes
        if event.button != 1 or (
            event.inaxes not in self.ui.lowerfigure.canvas.axes[0:2]
            and event.name == "button_press_event"
        ):
            return

        if event.name == "button_press_event":
            # button press sets the current axes, y location
            self.drag_axes = event.inaxes
            self.last_y = event.ydata
            self.ui.reset_tip_label.setText("")
        else:
            # display tip if y position is extreme
            if max([abs(self.eeg_signal_offset), abs(self.emg_signal_offset)]) > 0.8:
                self.ui.reset_tip_label.setText(RESET_TIP)

            # button release clears current axes, y location
            self.drag_axes = None
            self.last_y = None

        # double click resets plot to original conditions
        if event.dblclick:
            if event.inaxes == self.ui.lowerfigure.canvas.axes[0]:
                self.eeg_signal_offset = 0
                self.eeg_signal_scale_factor = 1
                self._refresh_eeg_line()
            else:
                self.emg_signal_offset = 0
                self.emg_signal_scale_factor = 1
                self._refresh_emg_line()
            self.ui.lowerfigure.canvas.draw()
            self.ui.reset_tip_label.setText("")

    def lower_plot_move(self, event):
        """Triggered when the mouse moves in the lower plots"""
        # only process move events where LMB is down
        # and the axes match the most recent click
        if event.button != 1 or event.inaxes != self.drag_axes:
            return

        # get movement distance on the y-axis
        y_diff = event.ydata - self.last_y
        # update mouse's most recent y position
        self.last_y = event.ydata

        # update y offset for the selected plot
        if event.inaxes == self.ui.lowerfigure.canvas.axes[0]:
            self.eeg_signal_offset += y_diff
            self._refresh_eeg_line()
        else:
            self.emg_signal_offset += y_diff
            self._refresh_emg_line()
        self.ui.lowerfigure.canvas.draw()


def convert_labels(labels: np.array, style: str) -> np.array:
    """Convert labels between "display" and "digit" formats

    It's useful to represent brain state labels in two ways:
    Digit format: this is how labels are represented in files. It matches the digit
        attribute of the BrainState class as well as the number pressed on the
        keyboard to set an epoch to that brain state.
    Display format: the y-axis value associated with a brain state when brain state
        labels are displayed as an image. This is also the index of the brain state
        in the colormap. Undefined epochs are mapped to 0, and digits are mapped to
        the numbers 1-10 in keyboard order (1234567890).

    :param labels: brain state labels
    :param style: target format for the output
    :return: formatted labels
    """
    if style == DISPLAY_FORMAT:
        # convert 0 to 10, undefined to 0
        labels = [i if i != 0 else 10 for i in labels]
        return np.array([i if i != UNDEFINED_LABEL else 0 for i in labels])
    elif style == DIGIT_FORMAT:
        # convert 0 to undefined, 10 to 0
        labels = [i if i != 0 else UNDEFINED_LABEL for i in labels]
        return np.array([i if i != 10 else 0 for i in labels])
    else:
        raise ValueError(f"style must be '{DISPLAY_FORMAT}' or '{DIGIT_FORMAT}'")


def create_label_img(labels: np.array, label_display_options: np.array) -> np.array:
    """Create an image to display brain state labels

    :param labels: brain state labels, in "display" format
    :param label_display_options: y-axis locations of valid brain state labels
    :return: brain state label image
    """
    # While there can be up to 10 valid brain states, it's possible that not all of them
    # are in use. We don't need to display brain states below and above the range of
    # valid brain states, since those rows would always be empty.
    smallest_display_label = np.min(label_display_options)
    # "background" of the image is white
    label_img = np.ones(
        [
            (np.max(label_display_options) - smallest_display_label + 1),
            len(labels),
            4,
        ]
    )
    for i, label in enumerate(labels):
        if label > 0:
            label_img[label - smallest_display_label, i, :] = LABEL_CMAP[label]
        else:
            # label is undefined
            label_img[:, i] = np.array([0, 0, 0, 1])
    return label_img


def create_confidence_img(confidence_scores: np.array) -> np.array:
    """Create an image to display confidence scores

    :param confidence_scores: confidence scores
    :return: confidence score image
    """
    if confidence_scores is None:
        return None

    confidence_img = np.ones([1, len(confidence_scores), 3])
    for i, c in enumerate(confidence_scores):
        confidence_img[0, i, 1:] = c
    return confidence_img


def create_upper_emg_signal(
    emg: np.array,
    sampling_rate: int | float,
    epoch_length: int | float,
    emg_filter: EMGFilter,
) -> np.array:
    """Calculate RMS of EMG for each epoch and apply a ceiling

    :param emg: EMG signal
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :param emg_filter: EMG filter parameters
    :return: processed EMG signal
    """
    emg_rms = get_emg_power(
        emg,
        sampling_rate,
        epoch_length,
        emg_filter,
    )
    return np.clip(
        emg_rms, np.percentile(emg_rms, 0.1), np.mean(emg_rms) + np.std(emg_rms) * 2.5
    )


def transform_eeg_emg(eeg: np.array, emg: np.array) -> (np.array, np.array):
    """Center and scale the EEG and EMG signals

    A heuristic approach to fitting the EEG and EMG signals in the plot.

    :param eeg: EEG signal
    :param emg: EMG signal
    :return: centered and scaled signals
    """
    eeg = eeg - np.mean(eeg)
    emg = emg - np.mean(emg)
    eeg = eeg / np.percentile(np.abs(eeg), 98) / 2.2
    emg = emg / np.percentile(np.abs(emg), 98) / 2.2
    return eeg, emg


def find_new_x_limits(
    direction: str | None,
    left_epoch: int,
    right_epoch: int,
    total_epochs: int,
    min_n_shown: int,
    selected_epoch: int,
) -> (int, int):
    """Calculate new plot x limits to allow zooming

    :param direction: in, out, reset, or None
    :param left_epoch: index of current leftmost epoch
    :param right_epoch: index of current rightmost epoch
    :param total_epochs: total number of epochs in the recording
    :param min_n_shown: minimum number of epochs to display
    :param selected_epoch: currently selected epoch
    """
    # number of epochs currently displayed in the upper plots
    current_n_shown = right_epoch - left_epoch + 1
    if direction == ZOOM_IN:
        # can't display fewer than the number of epochs in the lower plot
        new_n_shown = max([min_n_shown, round(current_n_shown * (1 - ZOOM_FACTOR))])
    elif direction == ZOOM_OUT:
        # can't display more than the total number of epochs
        new_n_shown = min([total_epochs, round(current_n_shown / (1 - ZOOM_FACTOR))])
    elif direction == ZOOM_RESET:
        left_epoch = 0
        right_epoch = total_epochs - 1
        return left_epoch, right_epoch
    else:  # just recalculating if min_n_shown has changed
        new_n_shown = int(
            np.clip(current_n_shown, a_min=min_n_shown, a_max=total_epochs)
        )

    # count epochs to show on either side of the selected epoch
    epochs_on_left_side = int(np.ceil((new_n_shown - 1) / 2))
    epochs_on_right_side = new_n_shown - epochs_on_left_side - 1
    if selected_epoch - epochs_on_left_side < 0:
        # can't go further left than 0
        left_epoch = 0
        right_epoch = new_n_shown - 1
    elif selected_epoch + epochs_on_right_side >= total_epochs:
        # can't go further right than the total number of epochs
        left_epoch = total_epochs - new_n_shown
        right_epoch = total_epochs - 1
    else:
        left_epoch = selected_epoch - epochs_on_left_side
        right_epoch = selected_epoch + epochs_on_right_side

    return left_epoch, right_epoch



"""File dialog helpers"""

import os

from PySide6.QtWidgets import QFileDialog, QWidget


def select_existing_file(parent: QWidget, title: str, file_filter: str) -> str | None:
    """Show dialog to select an existing file.

    :param parent: parent widget
    :param title: dialog window title
    :param file_filter: file type filter (e.g., "*.csv")
    :return: normalized path or None if cancelled
    """
    dialog = QFileDialog(parent)
    dialog.setWindowTitle(title)
    dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
    dialog.setViewMode(QFileDialog.ViewMode.Detail)
    dialog.setNameFilter(file_filter)

    if dialog.exec():
        return os.path.normpath(dialog.selectedFiles()[0])
    return None


def select_save_location(parent: QWidget, caption: str, file_filter: str) -> str | None:
    """Show dialog to choose save location.

    :param parent: parent widget
    :param caption: dialog window caption
    :param file_filter: file type filter (e.g., "*.csv")
    :return: normalized path or None if cancelled
    """
    filename, _ = QFileDialog.getSaveFileName(
        parent, caption=caption, filter=file_filter
    )
    if filename:
        return os.path.normpath(filename)
    return None



"""Settings tab manager"""

from dataclasses import dataclass
from functools import partial

from PySide6.QtCore import QObject
from PySide6.QtWidgets import QCheckBox, QDoubleSpinBox, QLineEdit

from accusleepy.brain_state_set import BrainState, BrainStateSet
from accusleepy.constants import (
    DEFAULT_BATCH_SIZE,
    DEFAULT_DELETE_TRAINING_IMAGES_STATE,
    DEFAULT_EMG_BP_LOWER,
    DEFAULT_EMG_BP_UPPER,
    DEFAULT_EMG_FILTER_ORDER,
    DEFAULT_LEARNING_RATE,
    DEFAULT_MOMENTUM,
    DEFAULT_TRAINING_EPOCHS,
    UNDEFINED_LABEL,
)
from accusleepy.fileio import (
    AccuSleePyConfig,
    EMGFilter,
    Hyperparameters,
    save_config,
)
from accusleepy.gui.primary_window import Ui_PrimaryWindow


@dataclass
class StateSettings:
    """Widgets for config settings for a brain state"""

    digit: int
    enabled_widget: QCheckBox
    name_widget: QLineEdit
    is_scored_widget: QCheckBox
    frequency_widget: QDoubleSpinBox


class SettingsWidget(QObject):
    """Manages settings tab UI and configuration data"""

    def __init__(
        self,
        ui: Ui_PrimaryWindow,
        config: AccuSleePyConfig,
        parent: QObject | None = None,
    ):
        super().__init__(parent)
        self._ui = ui

        # Store configuration values (only settings managed by the Settings tab)
        self._brain_state_set = config.brain_state_set
        self._emg_filter = config.emg_filter
        self._hyperparameters = config.hyperparameters
        self._default_epochs_to_show = config.epochs_to_show
        self._default_autoscroll_state = config.autoscroll_state
        self._delete_training_images = config.delete_training_images

        # Store default values for main tab settings (used to populate Settings tab UI)
        self._default_epoch_length = config.default_epoch_length
        self._default_overwrite_setting = config.overwrite_setting
        self._default_confidence_setting = config.save_confidence_setting
        self._default_min_bout_length = config.min_bout_length

        # Initialize settings widgets
        self._settings_widgets: dict[int, StateSettings] = {}
        self._initialize_settings_tab()

    @property
    def brain_state_set(self) -> BrainStateSet:
        """The current brain state set configuration (from saved config)"""
        return self._brain_state_set

    @property
    def emg_filter(self) -> EMGFilter:
        """EMG filter parameters (from saved config)"""
        return self._emg_filter

    @property
    def hyperparameters(self) -> Hyperparameters:
        """Model training hyperparameters"""
        return self._hyperparameters

    @property
    def delete_training_images(self) -> bool:
        """Whether to delete images after training"""
        return self._delete_training_images

    @property
    def default_epochs_to_show(self) -> int:
        """Default number of epochs to show in manual scoring"""
        return self._default_epochs_to_show

    @property
    def default_autoscroll_state(self) -> bool:
        """Default autoscroll state for manual scoring"""
        return self._default_autoscroll_state

    def _initialize_settings_tab(self) -> None:
        """Populate settings tab and assign its callbacks"""
        # Store dictionary that maps digits to rows of widgets
        self._settings_widgets = {
            1: StateSettings(
                digit=1,
                enabled_widget=self._ui.enable_state_1,
                name_widget=self._ui.state_name_1,
                is_scored_widget=self._ui.state_scored_1,
                frequency_widget=self._ui.state_frequency_1,
            ),
            2: StateSettings(
                digit=2,
                enabled_widget=self._ui.enable_state_2,
                name_widget=self._ui.state_name_2,
                is_scored_widget=self._ui.state_scored_2,
                frequency_widget=self._ui.state_frequency_2,
            ),
            3: StateSettings(
                digit=3,
                enabled_widget=self._ui.enable_state_3,
                name_widget=self._ui.state_name_3,
                is_scored_widget=self._ui.state_scored_3,
                frequency_widget=self._ui.state_frequency_3,
            ),
            4: StateSettings(
                digit=4,
                enabled_widget=self._ui.enable_state_4,
                name_widget=self._ui.state_name_4,
                is_scored_widget=self._ui.state_scored_4,
                frequency_widget=self._ui.state_frequency_4,
            ),
            5: StateSettings(
                digit=5,
                enabled_widget=self._ui.enable_state_5,
                name_widget=self._ui.state_name_5,
                is_scored_widget=self._ui.state_scored_5,
                frequency_widget=self._ui.state_frequency_5,
            ),
            6: StateSettings(
                digit=6,
                enabled_widget=self._ui.enable_state_6,
                name_widget=self._ui.state_name_6,
                is_scored_widget=self._ui.state_scored_6,
                frequency_widget=self._ui.state_frequency_6,
            ),
            7: StateSettings(
                digit=7,
                enabled_widget=self._ui.enable_state_7,
                name_widget=self._ui.state_name_7,
                is_scored_widget=self._ui.state_scored_7,
                frequency_widget=self._ui.state_frequency_7,
            ),
            8: StateSettings(
                digit=8,
                enabled_widget=self._ui.enable_state_8,
                name_widget=self._ui.state_name_8,
                is_scored_widget=self._ui.state_scored_8,
                frequency_widget=self._ui.state_frequency_8,
            ),
            9: StateSettings(
                digit=9,
                enabled_widget=self._ui.enable_state_9,
                name_widget=self._ui.state_name_9,
                is_scored_widget=self._ui.state_scored_9,
                frequency_widget=self._ui.state_frequency_9,
            ),
            0: StateSettings(
                digit=0,
                enabled_widget=self._ui.enable_state_0,
                name_widget=self._ui.state_name_0,
                is_scored_widget=self._ui.state_scored_0,
                frequency_widget=self._ui.state_frequency_0,
            ),
        }

        # Update widget state to display current config
        # UI defaults for main tab settings (shown in Settings tab for configuration)
        self._ui.default_epoch_input.setValue(self._default_epoch_length)
        self._ui.overwrite_default_checkbox.setChecked(self._default_overwrite_setting)
        self._ui.confidence_setting_checkbox.setChecked(
            self._default_confidence_setting
        )
        self._ui.default_min_bout_length_spinbox.setValue(self._default_min_bout_length)
        self._ui.epochs_to_show_spinbox.setValue(self._default_epochs_to_show)
        self._ui.autoscroll_checkbox.setChecked(self._default_autoscroll_state)
        # EMG filter
        self._ui.emg_order_spinbox.setValue(self._emg_filter.order)
        self._ui.bp_lower_spinbox.setValue(self._emg_filter.bp_lower)
        self._ui.bp_upper_spinbox.setValue(self._emg_filter.bp_upper)
        # Model training hyperparameters
        self._ui.batch_size_spinbox.setValue(self._hyperparameters.batch_size)
        self._ui.learning_rate_spinbox.setValue(self._hyperparameters.learning_rate)
        self._ui.momentum_spinbox.setValue(self._hyperparameters.momentum)
        self._ui.training_epochs_spinbox.setValue(self._hyperparameters.training_epochs)
        self._ui.delete_image_box.setChecked(self._delete_training_images)
        # Brain states
        states = {b.digit: b for b in self._brain_state_set.brain_states}
        for digit in range(10):
            if digit in states:
                self._settings_widgets[digit].enabled_widget.setChecked(True)
                self._settings_widgets[digit].name_widget.setText(states[digit].name)
                self._settings_widgets[digit].is_scored_widget.setChecked(
                    states[digit].is_scored
                )
                self._settings_widgets[digit].frequency_widget.setValue(
                    states[digit].frequency
                )
            else:
                self._settings_widgets[digit].enabled_widget.setChecked(False)
                self._settings_widgets[digit].name_widget.setEnabled(False)
                self._settings_widgets[digit].is_scored_widget.setEnabled(False)
                self._settings_widgets[digit].frequency_widget.setEnabled(False)

        # Set callbacks
        self._ui.default_epoch_input.valueChanged.connect(self.reset_status_message)
        self._ui.overwrite_default_checkbox.stateChanged.connect(
            self.reset_status_message
        )
        self._ui.confidence_setting_checkbox.stateChanged.connect(
            self.reset_status_message
        )
        self._ui.default_min_bout_length_spinbox.valueChanged.connect(
            self.reset_status_message
        )
        self._ui.epochs_to_show_spinbox.valueChanged.connect(self.reset_status_message)
        self._ui.autoscroll_checkbox.stateChanged.connect(self.reset_status_message)

        self._ui.emg_order_spinbox.valueChanged.connect(self.reset_status_message)
        self._ui.bp_lower_spinbox.valueChanged.connect(
            self._emg_filter_bp_lower_changed
        )
        self._ui.bp_upper_spinbox.valueChanged.connect(
            self._emg_filter_bp_upper_changed
        )
        self._ui.batch_size_spinbox.valueChanged.connect(self._hyperparameters_changed)
        self._ui.learning_rate_spinbox.valueChanged.connect(
            self._hyperparameters_changed
        )
        self._ui.momentum_spinbox.valueChanged.connect(self._hyperparameters_changed)
        self._ui.training_epochs_spinbox.valueChanged.connect(
            self._hyperparameters_changed
        )
        self._ui.delete_image_box.stateChanged.connect(self._hyperparameters_changed)
        for digit in range(10):
            state = self._settings_widgets[digit]
            state.enabled_widget.stateChanged.connect(
                partial(self._set_brain_state_enabled, digit)
            )
            state.name_widget.editingFinished.connect(self.check_validity)
            state.is_scored_widget.stateChanged.connect(
                partial(self._is_scored_changed, digit)
            )
            state.frequency_widget.valueChanged.connect(self.check_validity)

    def _set_brain_state_enabled(self, digit: int, _state: int) -> None:
        """Called when user clicks "enabled" checkbox

        :param digit: brain state digit
        :param _state: unused but mandatory
        """
        state = self._settings_widgets[digit]
        is_checked = state.enabled_widget.isChecked()
        for widget in [
            state.name_widget,
            state.is_scored_widget,
        ]:
            widget.setEnabled(is_checked)
        state.frequency_widget.setEnabled(
            is_checked and state.is_scored_widget.isChecked()
        )
        if not is_checked:
            state.name_widget.setText("")
            state.frequency_widget.setValue(0)
        self.check_validity()

    def _is_scored_changed(self, digit: int, _state: int) -> None:
        """Called when user sets whether a state is scored

        :param digit: brain state digit
        :param _state: unused but mandatory
        """
        state = self._settings_widgets[digit]
        is_checked = state.is_scored_widget.isChecked()
        state.frequency_widget.setEnabled(is_checked)
        if not is_checked:
            state.frequency_widget.setValue(0)
        self.check_validity()

    def _emg_filter_bp_lower_changed(self, _new_value: int | float) -> None:
        """Called when user modifies EMG filter lower cutoff

        Value is read from UI when config is saved. Triggers validation.
        """
        self.check_validity()

    def _emg_filter_bp_upper_changed(self, _new_value: int | float) -> None:
        """Called when user modifies EMG filter upper cutoff

        Value is read from UI when config is saved. Triggers validation.
        """
        self.check_validity()

    def _hyperparameters_changed(self, _new_value) -> None:
        """Called when user modifies model training hyperparameters"""
        # These are the only changes to the settings tab that take
        # effect immediately
        self._hyperparameters = Hyperparameters(
            batch_size=self._ui.batch_size_spinbox.value(),
            learning_rate=self._ui.learning_rate_spinbox.value(),
            momentum=self._ui.momentum_spinbox.value(),
            training_epochs=self._ui.training_epochs_spinbox.value(),
        )
        self._delete_training_images = self._ui.delete_image_box.isChecked()
        self._ui.save_config_status.setText("")

    def reset_status_message(self, _new_value=None) -> None:
        """Clear the message next to the 'save' button"""
        self._ui.save_config_status.setText("")

    def check_validity(self) -> str | None:
        """Check if brain state configuration on screen is valid

        :return: error message if invalid, None if valid
        """
        message = None

        # Strip whitespace from brain state names and update display
        for digit in range(10):
            state = self._settings_widgets[digit]
            current_name = state.name_widget.text()
            formatted_name = current_name.strip()
            if current_name != formatted_name:
                state.name_widget.setText(formatted_name)

        # Check if names are unique and frequencies add up to 1
        names = []
        frequencies = []
        for digit in range(10):
            state = self._settings_widgets[digit]
            if state.enabled_widget.isChecked():
                names.append(state.name_widget.text())
                frequencies.append(state.frequency_widget.value())
        if len(names) != len(set(names)):
            message = "Error: names must be unique"
        if sum(frequencies) != 1:
            message = "Error: sum(frequencies) != 1"

        # Check validity of EMG filter settings (read from UI, not saved state)
        bp_lower = self._ui.bp_lower_spinbox.value()
        bp_upper = self._ui.bp_upper_spinbox.value()
        if bp_lower >= bp_upper:
            message = "Error: EMG filter cutoff frequencies are invalid"

        if message is not None:
            self._ui.save_config_status.setText(message)
            self._ui.save_config_button.setEnabled(False)
            return message

        self._ui.save_config_button.setEnabled(True)
        self._ui.save_config_status.setText("")
        return None

    def save_config(self) -> None:
        """Save configuration to file"""
        # Check that configuration is valid
        error_message = self.check_validity()
        if error_message is not None:
            return

        # Build a BrainStateSet object from the current configuration
        brain_states = []
        for digit in range(10):
            state = self._settings_widgets[digit]
            if state.enabled_widget.isChecked():
                brain_states.append(
                    BrainState(
                        name=state.name_widget.text(),
                        digit=digit,
                        is_scored=state.is_scored_widget.isChecked(),
                        frequency=state.frequency_widget.value(),
                    )
                )

        # Update brain state set and EMG filter from UI
        # Note that this only occurs when a valid configuration is saved
        self._brain_state_set = BrainStateSet(brain_states, UNDEFINED_LABEL)
        self._emg_filter = EMGFilter(
            order=self._ui.emg_order_spinbox.value(),
            bp_lower=self._ui.bp_lower_spinbox.value(),
            bp_upper=self._ui.bp_upper_spinbox.value(),
        )

        # Save to file
        save_config(
            brain_state_set=self._brain_state_set,
            default_epoch_length=self._ui.default_epoch_input.value(),
            overwrite_setting=self._ui.overwrite_default_checkbox.isChecked(),
            save_confidence_setting=self._ui.confidence_setting_checkbox.isChecked(),
            min_bout_length=self._ui.default_min_bout_length_spinbox.value(),
            emg_filter=self._emg_filter,
            hyperparameters=self._hyperparameters,
            epochs_to_show=self._ui.epochs_to_show_spinbox.value(),
            autoscroll_state=self._ui.autoscroll_checkbox.isChecked(),
            delete_training_images=self._ui.delete_image_box.isChecked(),
        )
        self._ui.save_config_status.setText("configuration saved")

    def reset_emg_filter(self) -> None:
        """Reset EMG filter settings to defaults"""
        self._ui.emg_order_spinbox.setValue(DEFAULT_EMG_FILTER_ORDER)
        self._ui.bp_lower_spinbox.setValue(DEFAULT_EMG_BP_LOWER)
        self._ui.bp_upper_spinbox.setValue(DEFAULT_EMG_BP_UPPER)

    def reset_hyperparams(self) -> None:
        """Reset hyperparameters to defaults"""
        self._ui.batch_size_spinbox.setValue(DEFAULT_BATCH_SIZE)
        self._ui.learning_rate_spinbox.setValue(DEFAULT_LEARNING_RATE)
        self._ui.momentum_spinbox.setValue(DEFAULT_MOMENTUM)
        self._ui.training_epochs_spinbox.setValue(DEFAULT_TRAINING_EPOCHS)
        self._ui.delete_image_box.setChecked(DEFAULT_DELETE_TRAINING_IMAGES_STATE)



"""Service classes for orchestrating AccuSleePy operations.

Isolating certain functionality here, without any interaction
with UI state, makes it more testable.
"""

import datetime
import logging
import os
import shutil
from collections.abc import Callable
from dataclasses import dataclass, field

import numpy as np
import pandas as pd

from accusleepy.bouts import enforce_min_bout_length
from accusleepy.brain_state_set import BrainStateSet
from accusleepy.constants import (
    ANNOTATIONS_FILENAME,
    CALIBRATION_ANNOTATION_FILENAME,
    DEFAULT_MODEL_TYPE,
    MIN_EPOCHS_PER_STATE,
    MIXTURE_MEAN_COL,
    MIXTURE_SD_COL,
    UNDEFINED_LABEL,
)
from accusleepy.fileio import (
    EMGFilter,
    Hyperparameters,
    Recording,
    load_calibration_file,
    load_labels,
    load_recording,
    save_labels,
)
from accusleepy.models import SSANN
from accusleepy.signal_processing import (
    create_eeg_emg_image,
    create_training_images,
    get_mixture_values,
    resample_and_standardize,
)
from accusleepy.validation import check_label_validity

logger = logging.getLogger(__name__)


@dataclass
class ServiceResult:
    """Result of a service operation."""

    success: bool
    messages: list[str] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)
    error: str | None = None

    def report_to(self, callback: Callable[[str], None]) -> None:
        """Report all warnings, messages, and errors through a callback.

        :param callback: function to call with each message string
        """
        for warning in self.warnings:
            callback(f"WARNING: {warning}")
        for message in self.messages:
            callback(message)
        if not self.success and self.error:
            callback(f"ERROR: {self.error}")


def check_single_file_inputs(
    recording: Recording,
    epoch_length: int | float,
) -> str | None:
    """Check that a recording's inputs appear valid.

    This runs some basic tests for whether it will be possible to
    load and score a recording. If any test fails, we return an
    error message.

    :param recording: the recording to validate
    :param epoch_length: epoch length in seconds
    :return: error message, or None if valid
    """
    if epoch_length == 0:
        return "epoch length can't be 0"
    if recording.sampling_rate == 0:
        return "sampling rate can't be 0"
    if epoch_length > recording.sampling_rate:
        return "invalid epoch length or sampling rate"
    if recording.recording_file == "":
        return "no recording selected"
    if not os.path.isfile(recording.recording_file):
        return "recording file does not exist"
    if recording.label_file == "":
        return "no label file selected"
    return None


class TrainingService:
    """Service for training classification models."""

    def __init__(self, progress_callback: Callable[[str], None] | None = None) -> None:
        """Initialize the training service.

        :param progress_callback: optional callback for progress messages
        """
        self.progress_callback = progress_callback

    def _report_progress(self, message: str) -> None:
        """Report progress if callback is available."""
        if self.progress_callback:
            self.progress_callback(message)

    def train_model(
        self,
        recordings: list[Recording],
        epoch_length: int | float,
        epochs_per_img: int,
        model_type: str,
        calibrate: bool,
        calibration_fraction: float,
        brain_state_set: BrainStateSet,
        emg_filter: EMGFilter,
        hyperparameters: Hyperparameters,
        model_filename: str,
        temp_image_dir: str | None = None,
        delete_images: bool = True,
    ) -> ServiceResult:
        """Train a classification model.

        :param recordings: list of recordings to use for training
        :param epoch_length: epoch length in seconds
        :param epochs_per_img: number of epochs per training image
        :param model_type: type of model ('default' or 'real-time')
        :param calibrate: whether to calibrate the model
        :param calibration_fraction: fraction of data to use for calibration
        :param brain_state_set: set of brain state options
        :param emg_filter: EMG filter parameters
        :param hyperparameters: model training hyperparameters
        :param model_filename: path to save the trained model
        :param temp_image_dir: directory for training images (auto-generated if None)
        :param delete_images: whether to delete training images after training
        :return: ServiceResult with status and messages
        """
        result = ServiceResult(success=True)

        # Validate epochs_per_img for default model type
        if model_type == DEFAULT_MODEL_TYPE and epochs_per_img % 2 == 0:
            return ServiceResult(
                success=False,
                error=(
                    "For the default model type, number of epochs "
                    "per image must be an odd number."
                ),
            )

        # Validate each recording
        for recording in recordings:
            error_message = check_single_file_inputs(recording, epoch_length)
            if error_message:
                return ServiceResult(
                    success=False,
                    error=f"Recording {recording.name}: {error_message}",
                )

        # Create temp image directory if not provided
        if temp_image_dir is None:
            temp_image_dir = os.path.join(
                os.path.dirname(model_filename),
                "images_" + datetime.datetime.now().strftime("%Y%m%d%H%M"),
            )

        if os.path.exists(temp_image_dir):
            result.warnings.append("Training image folder exists, will be overwritten")
        os.makedirs(temp_image_dir, exist_ok=True)

        # Create training images
        self._report_progress("Creating training images")
        if not delete_images:
            result.messages.append(f"Creating training images in {temp_image_dir}")
        else:
            result.messages.append(
                f"Creating temporary folder of training images: {temp_image_dir}"
            )

        logger.info("Creating training images")
        failed_recordings, training_class_balance, had_zero_variance = (
            create_training_images(
                recordings=recordings,
                output_path=temp_image_dir,
                epoch_length=epoch_length,
                epochs_per_img=epochs_per_img,
                brain_state_set=brain_state_set,
                model_type=model_type,
                calibration_fraction=calibration_fraction,
                emg_filter=emg_filter,
            )
        )

        if had_zero_variance:
            result.warnings.append(
                "Some recordings contain features with zero variance. "
                "The EEG or EMG signal might be empty. If this is unexpected, "
                "please make sure the recording files are correctly formatted."
            )

        if len(failed_recordings) > 0:
            if len(failed_recordings) == len(recordings):
                # Cleanup before returning error
                if delete_images and os.path.exists(temp_image_dir):
                    shutil.rmtree(temp_image_dir)
                return ServiceResult(
                    success=False,
                    error="No recordings were valid!",
                    warnings=result.warnings,
                )
            else:
                result.warnings.append(
                    "The following recordings could not be loaded and will not "
                    f"be used for training: {', '.join([str(r) for r in failed_recordings])}. "
                    "More information might be available in the terminal."
                )

        # Train model
        self._report_progress("Training model")
        logger.info("Training model")

        from accusleepy.classification import create_dataloader, train_ssann
        from accusleepy.models import save_model
        from accusleepy.temperature_scaling import ModelWithTemperature

        model = train_ssann(
            annotations_file=os.path.join(temp_image_dir, ANNOTATIONS_FILENAME),
            img_dir=temp_image_dir,
            training_class_balance=training_class_balance,
            n_classes=brain_state_set.n_classes,
            hyperparameters=hyperparameters,
        )

        # Calibrate the model if requested
        if calibrate:
            calibration_annotation_file = os.path.join(
                temp_image_dir, CALIBRATION_ANNOTATION_FILENAME
            )
            calibration_dataloader = create_dataloader(
                annotations_file=calibration_annotation_file,
                img_dir=temp_image_dir,
                hyperparameters=hyperparameters,
            )
            model = ModelWithTemperature(model)
            logger.info("Calibrating model")
            model.set_temperature(calibration_dataloader)

        # Save model
        save_model(
            model=model,
            filename=model_filename,
            epoch_length=epoch_length,
            epochs_per_img=epochs_per_img,
            model_type=model_type,
            brain_state_set=brain_state_set,
            is_calibrated=calibrate,
        )

        # Optionally delete images
        if delete_images:
            logger.info("Cleaning up training image folder")
            shutil.rmtree(temp_image_dir)

        result.messages.append(f"Training complete. Saved model to {model_filename}")
        logger.info("Training complete")

        return result


@dataclass
class LoadedModel:
    """State for a loaded classification model."""

    model: SSANN | None = None
    epoch_length: int | float | None = None
    epochs_per_img: int | None = None


def score_recording_list(
    recordings: list[Recording],
    loaded_model: LoadedModel,
    epoch_length: int | float,
    only_overwrite_undefined: bool,
    save_confidence_scores: bool,
    min_bout_length: int | float,
    brain_state_set: BrainStateSet,
    emg_filter: EMGFilter,
) -> ServiceResult:
    """Score all recordings using a classification model.

    :param recordings: list of recordings to score
    :param loaded_model: loaded classification model and metadata
    :param epoch_length: epoch length in seconds
    :param only_overwrite_undefined: only overwrite epochs labeled as undefined
    :param save_confidence_scores: whether to save confidence scores
    :param min_bout_length: minimum bout length in seconds
    :param brain_state_set: set of brain state options
    :param emg_filter: EMG filter parameters
    :return: ServiceResult with status and messages
    """
    result = ServiceResult(success=True)

    # Validate model is loaded
    if loaded_model.model is None:
        return ServiceResult(
            success=False,
            error="No classification model file selected",
        )

    # Validate min_bout_length
    if min_bout_length < epoch_length:
        return ServiceResult(
            success=False,
            error="Minimum bout length must be >= epoch length",
        )

    # Validate model epoch length matches current
    if epoch_length != loaded_model.epoch_length:
        return ServiceResult(
            success=False,
            error=(
                f"Model was trained with an epoch length of "
                f"{loaded_model.epoch_length} seconds, but the current "
                f"epoch length setting is {epoch_length} seconds."
            ),
        )

    # Validate each recording
    for recording in recordings:
        error_message = check_single_file_inputs(recording, epoch_length)
        if error_message:
            return ServiceResult(
                success=False,
                error=f"Recording {recording.name}: {error_message}",
            )
        if recording.calibration_file == "":
            return ServiceResult(
                success=False,
                error=f"Recording {recording.name}: no calibration file selected",
            )

    from accusleepy.classification import score_recording

    any_zero_variance = False

    # Score each recording
    for recording in recordings:
        # Load EEG, EMG
        try:
            eeg, emg = load_recording(recording.recording_file)
            sampling_rate = recording.sampling_rate

            eeg, emg, sampling_rate = resample_and_standardize(
                eeg=eeg,
                emg=emg,
                sampling_rate=sampling_rate,
                epoch_length=epoch_length,
            )
        except Exception:
            logger.exception("Failed to load %s", recording.recording_file)
            result.warnings.append(
                f"Could not load recording {recording.name}. "
                "This recording will be skipped."
            )
            continue

        # Load labels
        label_file = recording.label_file
        if os.path.isfile(label_file):
            try:
                existing_labels, _ = load_labels(label_file)
            except Exception:
                logger.exception("Failed to load %s", label_file)
                result.warnings.append(
                    f"Could not load existing labels for recording "
                    f"{recording.name}. This recording will be skipped."
                )
                continue
            # Only check the length
            samples_per_epoch = sampling_rate * epoch_length
            epochs_in_recording = round(eeg.size / samples_per_epoch)
            if epochs_in_recording != existing_labels.size:
                result.warnings.append(
                    f"Existing labels for recording {recording.name} "
                    "do not match the recording length. "
                    "This recording will be skipped."
                )
                continue
        else:
            existing_labels = None

        # Load calibration data
        if not os.path.isfile(recording.calibration_file):
            result.warnings.append(
                f"Calibration file does not exist for recording "
                f"{recording.name}. This recording will be skipped."
            )
            continue
        try:
            mixture_means, mixture_sds = load_calibration_file(
                recording.calibration_file
            )
        except Exception:
            logger.exception("Failed to load %s", recording.calibration_file)
            result.warnings.append(
                f"Could not load calibration file for recording "
                f"{recording.name}. This recording will be skipped."
            )
            continue

        # Check if calibration data contains any 0-variance features
        if np.any(mixture_sds == 0):
            any_zero_variance = True

        labels, confidence_scores = score_recording(
            model=loaded_model.model,
            eeg=eeg,
            emg=emg,
            mixture_means=mixture_means,
            mixture_sds=mixture_sds,
            sampling_rate=sampling_rate,
            epoch_length=epoch_length,
            epochs_per_img=loaded_model.epochs_per_img,
            brain_state_set=brain_state_set,
            emg_filter=emg_filter,
        )

        # Overwrite as needed
        if existing_labels is not None and only_overwrite_undefined:
            labels[existing_labels != UNDEFINED_LABEL] = existing_labels[
                existing_labels != UNDEFINED_LABEL
            ]

        # Enforce minimum bout length
        labels = enforce_min_bout_length(
            labels=labels,
            epoch_length=epoch_length,
            min_bout_length=min_bout_length,
        )

        # Ignore confidence scores if desired
        if not save_confidence_scores:
            confidence_scores = None

        # Save results
        save_labels(
            labels=labels, filename=label_file, confidence_scores=confidence_scores
        )
        result.messages.append(
            f"Saved labels for recording {recording.name} to {label_file}"
        )

    if any_zero_variance:
        result.warnings.append(
            "One or more calibration files has 0 variance "
            "for some features. This could indicate that the EEG or "
            "EMG signal is empty in the recording used for calibration."
        )

    return result


def create_calibration(
    recording: Recording,
    epoch_length: int | float,
    brain_state_set: BrainStateSet,
    emg_filter: EMGFilter,
    output_filename: str,
) -> ServiceResult:
    """Create a calibration file for a recording.

    :param recording: the recording to create calibration for
    :param epoch_length: epoch length in seconds
    :param brain_state_set: set of brain state options
    :param emg_filter: EMG filter parameters
    :param output_filename: path to save the calibration file
    :return: ServiceResult with status and messages
    """
    result = ServiceResult(success=True)

    # Validate recording inputs
    error_message = check_single_file_inputs(recording, epoch_length)
    if error_message:
        return ServiceResult(success=False, error=error_message)

    # Load the recording
    try:
        eeg, emg = load_recording(recording.recording_file)
    except Exception:
        logger.exception("Failed to load %s", recording.recording_file)
        return ServiceResult(
            success=False,
            error=(
                "Could not load recording. "
                "Check user manual for formatting instructions."
            ),
        )

    sampling_rate = recording.sampling_rate
    eeg, emg, sampling_rate = resample_and_standardize(
        eeg=eeg,
        emg=emg,
        sampling_rate=sampling_rate,
        epoch_length=epoch_length,
    )

    # Load and validate labels
    label_file = recording.label_file
    if not os.path.isfile(label_file):
        return ServiceResult(
            success=False,
            error="Label file does not exist",
        )

    try:
        labels, _ = load_labels(label_file)
    except Exception:
        logger.exception("Failed to load %s", label_file)
        return ServiceResult(
            success=False,
            error=(
                "Could not load labels. Check user manual for formatting instructions."
            ),
        )

    label_error_message = check_label_validity(
        labels=labels,
        confidence_scores=None,
        samples_in_recording=eeg.size,
        sampling_rate=sampling_rate,
        epoch_length=epoch_length,
        brain_state_set=brain_state_set,
    )
    if label_error_message:
        return ServiceResult(success=False, error=label_error_message)

    # Check that each scored brain state has sufficient observations
    for brain_state in brain_state_set.brain_states:
        if brain_state.is_scored:
            count = np.sum(labels == brain_state.digit)
            if count < MIN_EPOCHS_PER_STATE:
                return ServiceResult(
                    success=False,
                    error=(
                        f"At least {MIN_EPOCHS_PER_STATE} labeled epochs "
                        f"per brain state are required for calibration. Only "
                        f"{count} '{brain_state.name}' epoch(s) found."
                    ),
                )

    # Create calibration file
    img = create_eeg_emg_image(eeg, emg, sampling_rate, epoch_length, emg_filter)
    mixture_means, mixture_sds = get_mixture_values(
        img=img,
        labels=brain_state_set.convert_digit_to_class(labels),
        brain_state_set=brain_state_set,
    )
    pd.DataFrame({MIXTURE_MEAN_COL: mixture_means, MIXTURE_SD_COL: mixture_sds}).to_csv(
        output_filename, index=False
    )

    result.messages.append(
        f"Created calibration file using recording {recording.name} "
        f"at {output_filename}"
    )

    if np.any(mixture_sds == 0):
        result.warnings.append(
            "One or more features derived from the data have "
            "zero variance. This could indicate that the EEG or "
            "EMG signal is empty."
        )

    return result



"""Temperature scaling for model confidence calibration."""

import logging

import numpy as np
import torch
from torch import nn, optim
from torch.nn import functional as F

logger = logging.getLogger(__name__)


class ModelWithTemperature(nn.Module):
    """
    A thin decorator, which wraps a model with temperature scaling
    model (nn.Module):
        A classification neural network
        NB: Output of the neural network should be the classification logits,
            NOT the softmax (or log softmax)!
    """

    def __init__(self, model):
        super().__init__()
        self.model = model
        # https://github.com/gpleiss/temperature_scaling/issues/20
        # for another approach, see https://github.com/gpleiss/temperature_scaling/issues/36
        self.model.eval()
        self.temperature = nn.Parameter(torch.ones(1) * 1.5)

    def forward(self, x):
        logits = self.model(x)
        return self.temperature_scale(logits)

    def temperature_scale(self, logits):
        """
        Perform temperature scaling on logits
        """
        # Expand temperature to match the size of logits
        temperature = self.temperature.unsqueeze(1).expand(
            logits.size(0), logits.size(1)
        )
        return logits / temperature

    # This function probably should live outside of this class, but whatever
    def set_temperature(self, valid_loader):
        """
        Tune the temperature of the model (using the validation set).
        We're going to set it to optimize NLL.
        valid_loader (DataLoader): validation set loader
        """
        if torch.accelerator.is_available():
            device = torch.accelerator.current_accelerator().type
        else:
            device = "cpu"

        # self.cuda()
        self.to(device)
        nll_criterion = nn.CrossEntropyLoss().to(device)  # .cuda()
        ece_criterion = _ECELoss().to(device)  # .cuda()

        # First: collect all the logits and labels for the validation set
        logits_list = []
        labels_list = []
        prediction_list = []
        with torch.no_grad():
            for x, label in valid_loader:
                x = x.to(device)  # .cuda()
                logits = self.model(x)
                logits_list.append(logits)
                labels_list.append(label)

                _, pred = torch.max(logits, 1)
                prediction_list.append(pred)
            logits = torch.cat(logits_list).to(device)  # .cuda()
            labels = torch.cat(labels_list).to(device)  # .cuda()
            predictions = torch.cat(prediction_list).to(device)

        # Calculate NLL and ECE before temperature scaling
        before_temperature_nll = nll_criterion(logits, labels).item()
        before_temperature_ece = ece_criterion(logits, labels).item()
        logger.info(
            "Before temperature - NLL: %.3f, ECE: %.3f",
            before_temperature_nll,
            before_temperature_ece,
        )

        # Next: optimize the temperature w.r.t. NLL
        # https://github.com/gpleiss/temperature_scaling/issues/34
        optimizer = optim.LBFGS([self.temperature], lr=0.01, max_iter=100)

        def eval():
            optimizer.zero_grad()
            loss = nll_criterion(self.temperature_scale(logits), labels)
            loss.backward()
            return loss

        optimizer.step(eval)

        # Calculate NLL and ECE after temperature scaling
        after_temperature_nll = nll_criterion(
            self.temperature_scale(logits), labels
        ).item()
        after_temperature_ece = ece_criterion(
            self.temperature_scale(logits), labels
        ).item()
        logger.info("Optimal temperature: %.3f", self.temperature.item())
        logger.info(
            "After temperature - NLL: %.3f, ECE: %.3f",
            after_temperature_nll,
            after_temperature_ece,
        )

        val_acc = round(
            100 * np.mean(labels.cpu().numpy() == predictions.cpu().numpy()), 2
        )
        logger.info("Validation accuracy: %s%%", val_acc)

        return self


class _ECELoss(nn.Module):
    """
    Calculates the Expected Calibration Error of a model.
    (This isn't necessary for temperature scaling, just a cool metric).

    The input to this loss is the logits of a model, NOT the softmax scores.

    This divides the confidence outputs into equally-sized interval bins.
    In each bin, we compute the confidence gap:

    bin_gap = | avg_confidence_in_bin - accuracy_in_bin |

    We then return a weighted average of the gaps, based on the number
    of samples in each bin

    See: Naeini, Mahdi Pakdaman, Gregory F. Cooper, and Milos Hauskrecht.
    "Obtaining Well Calibrated Probabilities Using Bayesian Binning." AAAI.
    2015.
    """

    def __init__(self, n_bins=15):
        """
        n_bins (int): number of confidence interval bins
        """
        super().__init__()
        bin_boundaries = torch.linspace(0, 1, n_bins + 1)
        self.bin_lowers = bin_boundaries[:-1]
        self.bin_uppers = bin_boundaries[1:]

    def forward(self, logits, labels):
        softmaxes = F.softmax(logits, dim=1)
        confidences, predictions = torch.max(softmaxes, 1)
        accuracies = predictions.eq(labels)

        ece = torch.zeros(1, device=logits.device)
        for bin_lower, bin_upper in zip(self.bin_lowers, self.bin_uppers, strict=True):
            # Calculated |confidence - accuracy| in each bin
            in_bin = confidences.gt(bin_lower.item()) * confidences.le(bin_upper.item())
            prop_in_bin = in_bin.float().mean()
            if prop_in_bin.item() > 0:
                accuracy_in_bin = accuracies[in_bin].float().mean()
                avg_confidence_in_bin = confidences[in_bin].mean()
                ece += torch.abs(avg_confidence_in_bin - accuracy_in_bin) * prop_in_bin

        return ece



"""Application-wide constants and default settings."""

import numpy as np

# probably don't change these unless you really need to
UNDEFINED_LABEL = -1  # can't be the same as a brain state's digit, must be an integer
# calibration file columns
MIXTURE_MEAN_COL = "mixture_mean"
MIXTURE_SD_COL = "mixture_sd"
# recording file columns
EEG_COL = "eeg"
EMG_COL = "emg"
# label file columns
BRAIN_STATE_COL = "brain_state"
CONFIDENCE_SCORE_COL = "confidence_score"
# max number of messages to store in main window message box
MESSAGE_BOX_MAX_DEPTH = 200
# clip mixture z-scores above and below this level
# in the matlab implementation, 4.5 was used
ABS_MAX_Z_SCORE = 3.5
# upper frequency limit when generating EEG spectrograms
SPECTROGRAM_UPPER_FREQ = 64
# minimum number of epochs per brain state needed to create
# a calibration file or use a recording for model training
MIN_EPOCHS_PER_STATE = 3


# very unlikely you will want to change values from here onwards
# config file names
DEFAULT_CONFIG_FILE = "default_config.json"
USER_CONFIG_FILE = "config.json"
# number of times to include the EMG power in a training image
EMG_COPIES = 9
# minimum spectrogram window length, in seconds
MIN_WINDOW_LEN = 5
# frequency above which to downsample EEG spectrograms
DOWNSAMPLING_START_FREQ = 20
# highest EEG frequency used as model input
UPPER_FREQ = 50
# height in pixels of each training image
IMAGE_HEIGHT = (
    len(np.arange(0, DOWNSAMPLING_START_FREQ, 1 / MIN_WINDOW_LEN))
    + len(np.arange(DOWNSAMPLING_START_FREQ, UPPER_FREQ, 2 / MIN_WINDOW_LEN))
    + EMG_COPIES
)

# classification model types
DEFAULT_MODEL_TYPE = "default"  # current epoch is centered
REAL_TIME_MODEL_TYPE = "real-time"  # current epoch on the right
# valid filetypes
RECORDING_FILE_TYPES = [".parquet", ".csv"]
LABEL_FILE_TYPE = ".csv"
CALIBRATION_FILE_TYPE = ".csv"
MODEL_FILE_TYPE = ".pth"
# annotation file columns
FILENAME_COL = "filename"
LABEL_COL = "label"
# recording list file header:
RECORDING_LIST_NAME = "recording_list"
RECORDING_LIST_FILE_TYPE = ".json"
# filename used to store info about training image datasets
ANNOTATIONS_FILENAME = "annotations.csv"
# filename for annotation file for the calibration set
CALIBRATION_ANNOTATION_FILENAME = "calibration_set.csv"

# config file keys
# ui setting keys
DEFAULT_EPOCH_LENGTH_KEY = "default_epoch_length"
DEFAULT_CONFIDENCE_SETTING_KEY = "save_confidence_setting"
DEFAULT_MIN_BOUT_LENGTH_KEY = "default_min_bout_length"
DEFAULT_OVERWRITE_KEY = "default_overwrite_setting"
# EMG filter parameters key
EMG_FILTER_KEY = "emg_filter"
# model training hyperparameters key
HYPERPARAMETERS_KEY = "hyperparameters"
EPOCHS_TO_SHOW_KEY = "epochs_to_show"
AUTOSCROLL_KEY = "autoscroll_state"
DELETE_TRAINING_IMAGES_KEY = "delete_training_images"

# default values
# default UI settings
DEFAULT_MIN_BOUT_LENGTH = 5.0
DEFAULT_CONFIDENCE_SETTING = True
DEFAULT_OVERWRITE_SETTING = False
# default EMG filter parameters (order, bandpass frequencies)
DEFAULT_EMG_FILTER_ORDER = 8
DEFAULT_EMG_BP_LOWER = 20
DEFAULT_EMG_BP_UPPER = 50
# default hyperparameters
DEFAULT_BATCH_SIZE = 64
DEFAULT_LEARNING_RATE = 1e-3
DEFAULT_MOMENTUM = 0.9
DEFAULT_TRAINING_EPOCHS = 6
DEFAULT_DELETE_TRAINING_IMAGES_STATE = True
# default manual scoring settings
DEFAULT_EPOCHS_TO_SHOW = 5
DEFAULT_AUTOSCROLL_STATE = False



"""Model training and brain state classification."""

import os

import numpy as np
import pandas as pd
import torch
import torch.optim as optim
from torch import nn
from torch.utils.data import DataLoader, Dataset
from torchvision.io import read_image
from tqdm import trange

import accusleepy.constants as c
from accusleepy.brain_state_set import BrainStateSet
from accusleepy.fileio import EMGFilter, Hyperparameters
from accusleepy.models import SSANN
from accusleepy.signal_processing import (
    create_eeg_emg_image,
    format_img,
    mixture_z_score_img,
)


class AccuSleepImageDataset(Dataset):
    """Dataset for loading AccuSleep training images"""

    def __init__(
        self, annotations_file, img_dir, transform=None, target_transform=None
    ):
        self.img_labels = pd.read_csv(annotations_file)
        self.img_dir = img_dir
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self):
        return len(self.img_labels)

    def __getitem__(self, idx):
        img_path = str(
            os.path.join(self.img_dir, self.img_labels.at[idx, c.FILENAME_COL])
        )
        image = read_image(img_path)
        label = self.img_labels.at[idx, c.LABEL_COL]
        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)
        return image, label


def get_device():
    """Get accelerator, if one is available"""
    return (
        torch.accelerator.current_accelerator().type
        if torch.accelerator.is_available()
        else "cpu"
    )


def create_dataloader(
    annotations_file: str,
    img_dir: str,
    hyperparameters: Hyperparameters,
    shuffle: bool = True,
) -> DataLoader:
    """Create DataLoader for a dataset of training or calibration images

    :param annotations_file: file with information on each training image
    :param img_dir: training image location
    :param hyperparameters: model training hyperparameters
    :param shuffle: reshuffle data for every epoch
    :return: DataLoader for the data

    """
    image_dataset = AccuSleepImageDataset(
        annotations_file=annotations_file,
        img_dir=img_dir,
    )
    return DataLoader(
        image_dataset, batch_size=hyperparameters.batch_size, shuffle=shuffle
    )


def train_ssann(
    annotations_file: str,
    img_dir: str,
    training_class_balance: np.ndarray,
    n_classes: int,
    hyperparameters: Hyperparameters,
) -> SSANN:
    """Train a SSANN classification model for sleep scoring

    :param annotations_file: file with information on each training image
    :param img_dir: training image location
    :param training_class_balance: proportion of each class in the training set
    :param n_classes: number of classes the model will learn
    :param hyperparameters: model training hyperparameters
    :return: trained Sleep Scoring Artificial Neural Network model
    """
    train_dataloader = create_dataloader(
        annotations_file=annotations_file,
        img_dir=img_dir,
        hyperparameters=hyperparameters,
    )

    device = get_device()
    model = SSANN(n_classes=n_classes)
    model.to(device)
    model.train()

    # correct for class imbalance
    weight = torch.tensor((training_class_balance**-1).astype("float32")).to(device)

    criterion = nn.CrossEntropyLoss(weight=weight)
    optimizer = optim.SGD(
        model.parameters(),
        lr=hyperparameters.learning_rate,
        momentum=hyperparameters.momentum,
    )

    for _ in trange(hyperparameters.training_epochs):
        for data in train_dataloader:
            inputs, labels = data
            (inputs, labels) = (inputs.to(device), labels.to(device))
            optimizer.zero_grad()
            outputs = model(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

    return model


def score_recording(
    model: SSANN,
    eeg: np.ndarray,
    emg: np.ndarray,
    mixture_means: np.ndarray,
    mixture_sds: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
    epochs_per_img: int,
    brain_state_set: BrainStateSet,
    emg_filter: EMGFilter,
) -> tuple[np.ndarray, np.ndarray]:
    """Use classification model to get brain state labels for a recording

    This assumes signals have been preprocessed to contain an integer
    number of epochs.

    :param model: classification model
    :param eeg: EEG signal
    :param emg: EMG signal
    :param mixture_means: mixture means, for calibration
    :param mixture_sds: mixture standard deviations, for calibration
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :param epochs_per_img: number of epochs for the model to consider
    :param brain_state_set: set of brain state options
    :param emg_filter: EMG filter parameters
    :return: brain state labels, confidence scores
    """
    # prepare model
    device = get_device()
    model = model.to(device)
    model.eval()

    # create and scale eeg+emg spectrogram
    img = create_eeg_emg_image(eeg, emg, sampling_rate, epoch_length, emg_filter)
    img, _ = mixture_z_score_img(
        img,
        mixture_means=mixture_means,
        mixture_sds=mixture_sds,
        brain_state_set=brain_state_set,
    )
    img = format_img(img=img, epochs_per_img=epochs_per_img, add_padding=True)

    # create dataset for inference
    images = []
    for i in range(img.shape[1] - epochs_per_img + 1):
        images.append(img[:, i : (i + epochs_per_img)].astype("float32"))
    images = torch.from_numpy(np.array(images))
    images = images[:, None, :, :]  # add channel
    images = images.to(device)

    # perform classification
    with torch.no_grad():
        outputs = model(images)
        logits, predicted = torch.max(outputs, 1)

    labels = brain_state_set.convert_class_to_digit(predicted.cpu().numpy())
    confidence_scores = 1 / (1 + np.e ** (-logits.cpu().numpy()))

    return labels, confidence_scores



"""Matplotlib FigureCanvas widget for manual scoring."""

from collections.abc import Callable

import matplotlib.axes
import matplotlib.ticker as mticker
import numpy as np
from matplotlib.backend_bases import MouseButton
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle
from matplotlib.widgets import RectangleSelector
from PySide6 import QtWidgets

from accusleepy.brain_state_set import BrainStateSet

# upper limit of spectrogram y-axis, in Hz
SPEC_UPPER_F = 30
# interval of spectrogram y-axis ticks, in Hz
SPEC_Y_TICK_INTERVAL = 10

# margins around subplots in the figure
SUBPLOT_TOP_MARGIN = 0.98
SUBPLOT_BOTTOM_MARGIN = 0.02
SUBPLOT_LEFT_MARGIN = 0.07
SUBPLOT_RIGHT_MARGIN = 0.95

# maximum number of x-axis ticks to show on the lower plot
MAX_LOWER_X_TICK_N = 7


class MplWidget(QtWidgets.QWidget):
    """Widget that displays a matplotlib FigureCanvas"""

    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self, parent)
        # set up the canvas and store a reference to its axes
        self.canvas = FigureCanvas(Figure())
        self.canvas.axes = None

        # set the widget layout and remove the margins
        vertical_layout = QtWidgets.QVBoxLayout()
        vertical_layout.addWidget(self.canvas)
        vertical_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vertical_layout)

        # given during the setup process
        self.epoch_length = None

        # upper plot references
        self.upper_marker = None
        self.label_img_ref = None
        self.spec_ref = None
        self.roi = None
        self.editing_patch = None
        self.roi_patch = None

        # lower plot references
        self.eeg_line = None
        self.emg_line = None
        self.top_marker = None
        self.bottom_marker = None
        self.eeg_epoch_plot = None
        self.emg_epoch_plot = None

    def setup_upper_figure(
        self,
        n_epochs: int,
        label_img: np.array,
        confidence_scores: np.array,
        confidence_img: np.array,
        spec: np.array,
        f: np.array,
        emg: np.array,
        epochs_to_show: int,
        label_display_options: np.array,
        brain_state_set: BrainStateSet,
        roi_function: Callable,
    ):
        """Initialize upper FigureCanvas for the manual scoring GUI

        :param n_epochs: number of epochs in the recording
        :param label_img: brain state labels, as an image
        :param confidence_scores: confidence scores
        :param confidence_img: confidence scores, as an image
        :param spec: EEG spectrogram image
        :param f: EEG spectrogram frequency axis
        :param emg: EMG RMS per epoch
        :param epochs_to_show: number of epochs to show in the lower plot
        :param label_display_options: valid brain state y-axis locations
        :param brain_state_set: set of brain states options
        :param roi_function: callback for ROI selection
        """
        # references to parts of the epoch marker
        self.upper_marker = list()

        # subplot layout
        height_ratios = [2, 8, 2, 12, 13]
        gs1 = GridSpec(5, 1, hspace=0.02, height_ratios=height_ratios)
        gs2 = GridSpec(5, 1, hspace=0.4, height_ratios=height_ratios)
        axes = list()
        axes.append(self.canvas.figure.add_subplot(gs1[0]))
        axes.append(self.canvas.figure.add_subplot(gs1[1]))
        axes.append(self.canvas.figure.add_subplot(gs1[2]))
        axes.append(self.canvas.figure.add_subplot(gs1[3]))
        axes.append(self.canvas.figure.add_subplot(gs2[4]))

        # subplots have different axes limits
        for i in [0, 1, 2, 4]:
            axes[i].set_xlim((-0.5, n_epochs - 0.5))
        axes[3].set_xlim(0, n_epochs)

        # confidence score subplot
        if confidence_scores is None:
            axes[0].set_visible(False)
        else:
            axes[0].set_ylim([-0.5, 0.5])
            axes[0].set_xticks([])
            axes[0].set_yticks([0])
            axes[0].set_yticklabels(["Conf."])
            axes[0].tick_params(axis="y", color="white")
            axes[0].imshow(
                confidence_img, aspect="auto", origin="lower", interpolation="None"
            )
            confidence_x = (
                np.repeat(list(range(len(confidence_scores) + 1)), 2)[1:-1] - 0.5
            )
            confidence_y = np.repeat(confidence_scores, 2) - 0.5
            axes[0].plot(confidence_x, confidence_y, "k", linewidth=0.5)
            for side in ["left", "right", "bottom", "top"]:
                axes[0].spines[side].set_visible(False)

        # brain state subplot
        axes[1].set_ylim(
            [-0.5, np.max(label_display_options) - np.min(label_display_options) + 0.5]
        )
        axes[1].set_xticks([])
        axes[1].set_yticks(
            label_display_options - np.min(label_display_options),
        )
        axes[1].set_yticklabels([b.name for b in brain_state_set.brain_states])
        ax2 = axes[1].secondary_yaxis("right")
        ax2.set_yticks(
            label_display_options - np.min(label_display_options),
        )
        ax2.set_yticklabels([b.digit for b in brain_state_set.brain_states])
        self.label_img_ref = axes[1].imshow(
            label_img, aspect="auto", origin="lower", interpolation="None"
        )
        # add patch to dim the display when creating an ROI
        self.editing_patch = axes[1].add_patch(
            Rectangle(
                xy=(-0.5, -0.5),
                width=n_epochs,
                height=np.max(label_display_options)
                - np.min(label_display_options)
                + 1,
                color="white",
                edgecolor=None,
                alpha=0.4,
                visible=False,
            )
        )
        # add the ROI selection widget, but disable it until it's needed
        self.roi = RectangleSelector(
            ax=axes[1],
            onselect=roi_function,
            interactive=False,
            button=MouseButton(1),
        )
        self.roi.set_active(False)
        # keep a reference to the ROI patch so we can change its color later
        # index 0 is the "editing_patch" created earlier
        self.roi_patch = [c for c in axes[1].get_children() if type(c) is Rectangle][1]

        # epoch marker subplot
        axes[2].set_ylim((0, 1))
        axes[2].axis("off")
        self.upper_marker.append(
            axes[2].plot([-0.5, epochs_to_show - 0.5], [0.5, 0.5], "r")[0]
        )
        self.upper_marker.append(axes[2].plot([0], [0.5], "rD")[0])

        # EEG spectrogram subplot
        # select subset of frequencies to show
        f = f[f <= SPEC_UPPER_F]
        spec = spec[0 : len(f), :]
        axes[3].set_ylabel("EEG", rotation="horizontal", ha="right")
        axes[3].set_yticks(
            np.linspace(
                0,
                len(f),
                1 + round(SPEC_UPPER_F / SPEC_Y_TICK_INTERVAL),
            ),
        )
        axes[3].set_yticklabels(
            np.arange(0, SPEC_UPPER_F + SPEC_Y_TICK_INTERVAL, SPEC_Y_TICK_INTERVAL)
        )
        axes[3].tick_params(axis="both", which="major", labelsize=8)
        axes[3].xaxis.set_major_formatter(mticker.FuncFormatter(self.time_formatter))
        self.spec_ref = axes[3].imshow(
            spec,
            vmin=np.percentile(spec, 2),
            vmax=np.percentile(spec, 98),
            aspect="auto",
            origin="lower",
            interpolation="None",
            extent=(
                0,
                n_epochs,
                -0.5,
                len(f) + 0.5,
            ),
        )

        # EMG subplot
        axes[4].set_xticks([])
        axes[4].set_yticks([])
        axes[4].set_ylabel("EMG", rotation="horizontal", ha="right")
        axes[4].plot(
            emg,
            "k",
            linewidth=0.5,
        )

        self.canvas.figure.subplots_adjust(
            left=SUBPLOT_LEFT_MARGIN,
            right=SUBPLOT_RIGHT_MARGIN,
            top=SUBPLOT_TOP_MARGIN,
            bottom=SUBPLOT_BOTTOM_MARGIN,
        )

        self.canvas.axes = axes

    def setup_lower_figure(
        self,
        label_img: np.array,
        sampling_rate: int | float,
        epochs_to_show: int,
        brain_state_set: BrainStateSet,
        label_display_options: np.array,
    ):
        """Initialize lower FigureCanvas for the manual scoring GUI

        :param label_img: brain state labels, as an image
        :param sampling_rate: EEG/EMG sampling rate, in Hz
        :param epochs_to_show: number of epochs to show in the lower plot
        :param brain_state_set: set of brain states options
        :param label_display_options: valid brain state y-axis locations
        """
        # number of samples in one epoch
        # we can expect this to be very close to an integer
        samples_per_epoch = round(sampling_rate * self.epoch_length)
        # number of EEG/EMG samples to plot
        samples_shown = round(samples_per_epoch * epochs_to_show)

        # references to parts of the epoch markers
        self.top_marker = list()
        self.bottom_marker = list()
        # epoch marker display parameters
        marker_dy = 0.25
        marker_y_offset_top = 0.02
        marker_y_offset_bottom = 0.01

        # subplot layout
        gs1 = GridSpec(3, 1, hspace=0)
        gs2 = GridSpec(3, 1, hspace=0.5)
        axes = list()
        axes.append(self.canvas.figure.add_subplot(gs1[0]))
        axes.append(self.canvas.figure.add_subplot(gs1[1]))
        axes.append(self.canvas.figure.add_subplot(gs2[2]))

        # EEG subplot
        axes[0].set_xticks([])
        axes[0].set_yticks([])
        axes[0].set_xlim((0, samples_shown))
        axes[0].set_ylim((-1, 1))
        axes[0].set_ylabel("EEG", rotation="horizontal", ha="right")
        self.eeg_line = axes[0].plot(
            np.zeros(samples_shown),
            "k",
            linewidth=0.5,
        )[0]
        self.eeg_epoch_plot = plot_eeg_emg_epochs(
            ax=axes[0],
            samples_per_epoch=samples_per_epoch,
            samples_shown=samples_shown,
            epochs_to_show=epochs_to_show,
        )
        self.eeg_epoch_plot.set(visible=False)

        # top epoch marker
        marker_x = [
            [0, 0],
            [0, samples_per_epoch],
            [samples_per_epoch, samples_per_epoch],
        ]
        marker_y = np.array(
            [
                [1 - marker_dy, 1],
                [1, 1],
                [1 - marker_dy, 1],
            ]
        )
        for x, y in zip(marker_x, marker_y, strict=True):
            self.top_marker.append(axes[0].plot(x, y - marker_y_offset_top, "r")[0])

        # EMG subplot
        axes[1].set_xticks(
            resample_x_ticks(
                np.arange(
                    0,
                    samples_shown,
                    samples_per_epoch,
                )
            )
        )
        axes[1].tick_params(axis="x", which="major", labelsize=8)
        axes[1].set_yticks([])
        axes[1].set_xlim((0, samples_shown))
        axes[1].set_ylim((-1, 1))
        axes[1].set_ylabel("EMG", rotation="horizontal", ha="right")

        self.emg_epoch_plot = plot_eeg_emg_epochs(
            ax=axes[1],
            samples_per_epoch=samples_per_epoch,
            samples_shown=samples_shown,
            epochs_to_show=epochs_to_show,
        )
        self.emg_epoch_plot.set(visible=False)

        self.emg_line = axes[1].plot(
            np.zeros(samples_shown),
            "k",
            linewidth=0.5,
        )[0]

        for x, y in zip(marker_x, marker_y, strict=True):
            self.bottom_marker.append(
                axes[1].plot(x, -1 * (y - marker_y_offset_bottom), "r")[0]
            )

        # brain state subplot
        axes[2].set_xticks([])
        axes[2].set_yticks(
            label_display_options - np.min(label_display_options),
        )
        axes[2].set_yticklabels([b.name for b in brain_state_set.brain_states])
        ax2 = axes[2].secondary_yaxis("right")
        ax2.set_yticks(
            label_display_options - np.min(label_display_options),
        )
        ax2.set_yticklabels([b.digit for b in brain_state_set.brain_states])
        axes[2].set_xlim((-0.5, epochs_to_show - 0.5))
        axes[2].set_ylim(
            [-0.5, np.max(label_display_options) - np.min(label_display_options) + 0.5]
        )
        self.label_img_ref = axes[2].imshow(
            label_img[:, 0:epochs_to_show, :],
            aspect="auto",
            origin="lower",
            interpolation="None",
        )

        self.canvas.figure.subplots_adjust(
            left=SUBPLOT_LEFT_MARGIN,
            right=SUBPLOT_RIGHT_MARGIN,
            top=SUBPLOT_TOP_MARGIN,
            bottom=SUBPLOT_BOTTOM_MARGIN,
        )

        self.canvas.axes = axes

    def time_formatter(self, x, pos):
        x = x * self.epoch_length
        return f"{int(x // 3600):02d}:{int(x // 60) % 60:02d}:{x % 60:05.2f}"


def resample_x_ticks(x_ticks: np.array) -> np.array:
    """Choose a subset of x_ticks to display

    The x-axis can get crowded if there are too many timestamps shown.
    This function finds a subset of evenly spaced x-axis ticks that
    includes the one at the beginning of the central epoch.

    :param x_ticks: full set of x_ticks
    :return: smaller subset of x_ticks
    """
    if len(x_ticks) <= MAX_LOWER_X_TICK_N:
        return x_ticks

    # number of ticks to the left of the central epoch
    # this will always be an integer
    nl = round((len(x_ticks) - 1) / 2)

    # search for even tick spacings that include the central epoch
    # if necessary, skip the leftmost tick
    for offset in [0, 1]:
        if (nl - offset) % 3 == 0:
            return x_ticks[offset :: round((nl - offset) / 3)]
        elif (nl - offset) % 2 == 0:
            return x_ticks[offset :: round((nl - offset) / 2)]


def plot_eeg_emg_epochs(
    ax: matplotlib.axes, samples_per_epoch: int, samples_shown: int, epochs_to_show: int
):
    """Plot epoch boundary markers on EEG/EMG data

    :param ax: axes on which to plot
    :param samples_per_epoch: number of samples in each epoch
    :param samples_shown: number of samples of data displayed
    :param epochs_to_show: number of epochs displayed
    :return: handle to plot object
    """
    return ax.plot(
        np.repeat(np.arange(samples_per_epoch, samples_shown, samples_per_epoch), 2),
        np.tile(np.array([-1.5, 1.5, 1.5, -1.5]), epochs_to_show)[
            : epochs_to_show * 2 - 2
        ],
        "gray",
        linewidth=1,
    )[0]
